#!/usr/bin/env python3
"""
01. 基本的な使用方法 - EphemeralDBの基本操作を学ぶ
"""

from ephemeraldb import EphemeralDB


def basic_operations():
    """基本的なCRUD操作のデモンストレーション"""
    print("=== 基本操作 ===")
    
    db = EphemeralDB()
    
    # 様々なデータ型の保存
    db.set('名前', 'アリス')
    db.set('年齢', 30)
    db.set('設定', {'テーマ': '暗い', '言語': '日本語'})
    db.set('スコア', [85, 92, 78])
    
    # データの取得
    print(f"名前: {db.get('名前')}")
    print(f"年齢: {db.get('年齢')}")
    print(f"設定: {db.get('設定')}")
    print(f"スコア: {db.get('スコア')}")
    
    # 存在確認
    print(f"'名前'が存在: {db.exists('名前')}")
    print(f"'不明'が存在: {db.exists('不明')}")
    
    # データの削除
    db.delete('年齢')
    print(f"削除後の年齢: {db.get('年齢')}")
    
    print()


def nested_keys():
    """ネストしたキー操作のデモンストレーション"""
    print("=== ネストしたキー ===")
    
    db = EphemeralDB()
    
    # ドット記法でネストしたデータを設定
    db.set('ユーザー.プロフィール.名前', 'ボブ')
    db.set('ユーザー.プロフィール.メール', 'bob@example.com')
    db.set('ユーザー.設定.テーマ', '明い')
    db.set('ユーザー.設定.通知', True)
    
    # ネストしたデータにアクセス
    print(f"ユーザー名: {db.get('ユーザー.プロフィール.名前')}")
    print(f"メール: {db.get('ユーザー.プロフィール.メール')}")
    print(f"テーマ: {db.get('ユーザー.設定.テーマ')}")
    print(f"通知: {db.get('ユーザー.設定.通知')}")
    
    # 全キーのリスト
    print(f"全キー: {db.keys()}")
    
    print()


def default_values():
    """デフォルト値の使用例"""
    print("=== デフォルト値 ===")
    
    db = EphemeralDB()
    
    # 存在しないキーにデフォルト値を設定
    print(f"存在しないキー: {db.get('存在しない')}")
    print(f"デフォルト値付き: {db.get('存在しない', 'デフォルト')}")
    print(f"数値デフォルト: {db.get('存在しない', 42)}")
    
    # 設定されたキーはデフォルト値を無視
    db.set('設定済み', '実際の値')
    print(f"設定済みキー: {db.get('設定済み', 'デフォルト')}")
    
    print()


def data_types():
    """様々なデータ型の保存と取得"""
    print("=== データ型 ===")
    
    db = EphemeralDB()
    
    # 様々なデータ型をテスト
    test_data = {
        '文字列': 'これは文字列です',
        '整数': 42,
        '浮動小数点': 3.14159,
        'ブール値': True,
        'リスト': [1, 2, 3, '混合', True],
        '辞書': {
            'ネスト': {
                '深い': 'ネスト構造',
                '数値': 100
            }
        },
        'タプル': (1, 2, 3),
        'None値': None
    }
    
    # データを保存
    for key, value in test_data.items():
        db.set(key, value)
    
    # データを取得して型を確認
    for key in test_data.keys():
        value = db.get(key)
        print(f"{key}: {value} (型: {type(value).__name__})")
    
    print()


if __name__ == "__main__":
    basic_operations()
    nested_keys()
    default_values()
    data_types()
    
    print("=== まとめ ===")
    print("EphemeralDBは様々なデータ型を保存でき、")
    print("ドット記法でネストしたキーもサポートします。")
    print("デフォルト値を使用して安全にデータにアクセスできます。")


#!/usr/bin/env python3
"""
06. テンプレートエンジン - 変数スコープを持つテンプレート処理
"""

from ephemeraldb import EphemeralDB
import re
from typing import Dict, Any, List, Optional


class SimpleTemplateEngine:
    """EphemeralDBを使用したシンプルなテンプレートエンジン"""
    
    def __init__(self):
        self.db = EphemeralDB()
        self.includes = {}  # インクルードされるテンプレート
        self.filters = {
            'upper': str.upper,
            'lower': str.lower,
            'title': str.title,
            'length': len,
            'reverse': lambda x: x[::-1] if isinstance(x, str) else str(x)[::-1],
        }
    
    def set_global_variable(self, key: str, value: Any):
        """グローバル変数を設定"""
        self.db.set(key, value)
    
    def register_include(self, name: str, template: str):
        """インクルード用テンプレートを登録"""
        self.includes[name] = template
    
    def render(self, template: str, variables: Optional[Dict[str, Any]] = None) -> str:
        """テンプレートをレンダリング"""
        with self.db.scope('レンダリング'):
            # ローカル変数を設定
            if variables:
                for key, value in variables.items():
                    self.db.set(key, value)
            
            # テンプレートを処理
            return self._process_template(template)
    
    def render_with_layout(self, layout: str, content: str, variables: Optional[Dict[str, Any]] = None) -> str:
        """レイアウト付きレンダリング"""
        with self.db.scope('レイアウト'):
            # コンテンツを変数として設定
            self.db.set('content', content)
            
            if variables:
                for key, value in variables.items():
                    self.db.set(key, value)
            
            return self._process_template(layout)
    
    def _process_template(self, template: str) -> str:
        """テンプレートを処理"""
        result = template
        
        # 1. インクルード処理 {{ include "template_name" }}
        result = self._process_includes(result)
        
        # 2. 変数置換 {{ variable_name }}
        result = self._process_variables(result)
        
        # 3. フィルター処理 {{ variable_name | filter }}
        result = self._process_filters(result)
        
        # 4. 条件分岐 {% if condition %} ... {% endif %}
        result = self._process_conditionals(result)
        
        # 5. ループ処理 {% for item in list %} ... {% endfor %}
        result = self._process_loops(result)
        
        return result
    
    def _process_includes(self, template: str) -> str:
        """インクルード処理"""
        include_pattern = r'\{\{\s*include\s+"([^"]+)"\s*\}\}'
        
        def replace_include(match):
            include_name = match.group(1)
            if include_name in self.includes:
                # インクルードを新しいスコープで処理
                with self.db.scope(f'インクルード_{include_name}'):
                    return self._process_template(self.includes[include_name])
            return f"[ERROR: インクルード '{include_name}' が見つかりません]"
        
        return re.sub(include_pattern, replace_include, template)
    
    def _process_variables(self, template: str) -> str:
        """変数置換処理"""
        var_pattern = r'\{\{\s*([a-zA-Z_][a-zA-Z0-9_.]*)\s*\}\}'
        
        def replace_var(match):
            var_name = match.group(1)
            value = self.db.get(var_name)
            return str(value) if value is not None else f"[UNDEFINED: {var_name}]"
        
        return re.sub(var_pattern, replace_var, template)
    
    def _process_filters(self, template: str) -> str:
        """フィルター処理"""
        filter_pattern = r'\{\{\s*([a-zA-Z_][a-zA-Z0-9_.]*)\s*\|\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*\}\}'
        
        def apply_filter(match):
            var_name = match.group(1)
            filter_name = match.group(2)
            
            value = self.db.get(var_name)
            if value is None:
                return f"[UNDEFINED: {var_name}]"
            
            if filter_name in self.filters:
                try:
                    filtered_value = self.filters[filter_name](value)
                    return str(filtered_value)
                except Exception as e:
                    return f"[FILTER ERROR: {filter_name} - {e}]"
            else:
                return f"[UNKNOWN FILTER: {filter_name}]"
        
        return re.sub(filter_pattern, apply_filter, template)
    
    def _process_conditionals(self, template: str) -> str:
        """条件分岐処理"""
        if_pattern = r'\{%\s*if\s+([a-zA-Z_][a-zA-Z0-9_.]*)\s*%\}(.*?)\{%\s*endif\s*%\}'
        
        def process_if(match):
            var_name = match.group(1)
            content = match.group(2)
            
            value = self.db.get(var_name)
            if value:  # Truthy check
                return self._process_template(content)
            else:
                return ""
        
        return re.sub(if_pattern, process_if, template, flags=re.DOTALL)
    
    def _process_loops(self, template: str) -> str:
        """ループ処理"""
        for_pattern = r'\{%\s*for\s+([a-zA-Z_][a-zA-Z0-9_]*)\s+in\s+([a-zA-Z_][a-zA-Z0-9_.]*)\s*%\}(.*?)\{%\s*endfor\s*%\}'
        
        def process_for(match):
            item_var = match.group(1)
            list_var = match.group(2)
            content = match.group(3)
            
            items = self.db.get(list_var)
            if not isinstance(items, (list, tuple)):
                return f"[ERROR: {list_var} はリストではありません]"
            
            result = []
            for i, item in enumerate(items):
                # 各アイテムを新しいスコープで処理
                with self.db.scope(f'ループ_{i}'):
                    self.db.set(item_var, item)
                    self.db.set('loop_index', i)
                    self.db.set('loop_first', i == 0)
                    self.db.set('loop_last', i == len(items) - 1)
                    result.append(self._process_template(content))
            
            return ''.join(result)
        
        return re.sub(for_pattern, process_for, template, flags=re.DOTALL)


def basic_template_demo():
    """基本的なテンプレート機能のデモ"""
    print("=== 基本テンプレート機能 ===")
    
    engine = SimpleTemplateEngine()
    
    # グローバル変数設定
    engine.set_global_variable('サイト名', 'マイウェブサイト')
    engine.set_global_variable('年', 2024)
    
    # 基本的な変数置換
    template1 = """
こんにちは、{{ ユーザー名 }}さん！
{{ サイト名 }}へようこそ。
今年は{{ 年 }}年です。
"""
    
    result1 = engine.render(template1, {
        'ユーザー名': 'アリス'
    })
    
    print("基本テンプレート結果:")
    print(result1)


def filter_demo():
    """フィルター機能のデモ"""
    print("=== フィルター機能 ===")
    
    engine = SimpleTemplateEngine()
    
    template = """
元の名前: {{ 名前 }}
大文字: {{ 名前 | upper }}
小文字: {{ 名前 | lower }}
タイトルケース: {{ 名前 | title }}
逆順: {{ 名前 | reverse }}
文字数: {{ 名前 | length }}
"""
    
    result = engine.render(template, {
        '名前': 'Hello World'
    })
    
    print("フィルター結果:")
    print(result)


def conditional_demo():
    """条件分岐のデモ"""
    print("=== 条件分岐 ===")
    
    engine = SimpleTemplateEngine()
    
    template = """
ユーザー: {{ ユーザー名 }}
{% if 管理者 %}
管理者権限があります。
管理者メニューを表示します。
{% endif %}
{% if プレミアム %}
プレミアム機能が利用可能です。
{% endif %}
"""
    
    # 管理者ユーザー
    result1 = engine.render(template, {
        'ユーザー名': 'admin',
        '管理者': True,
        'プレミアム': False
    })
    
    print("管理者ユーザー:")
    print(result1)
    
    # 一般ユーザー
    result2 = engine.render(template, {
        'ユーザー名': 'user',
        '管理者': False,
        'プレミアム': True
    })
    
    print("一般ユーザー:")
    print(result2)


def loop_demo():
    """ループ処理のデモ"""
    print("=== ループ処理 ===")
    
    engine = SimpleTemplateEngine()
    
    template = """
商品一覧:
{% for 商品 in 商品リスト %}
  {{ loop_index }}. {{ 商品.名前 }} - ¥{{ 商品.価格 }}
  {% if loop_first %}(新着商品){% endif %}
  {% if loop_last %}(最後の商品){% endif %}
{% endfor %}

合計: {{ 商品リスト | length }}件
"""
    
    result = engine.render(template, {
        '商品リスト': [
            {'名前': 'ノートPC', '価格': 89800},
            {'名前': 'マウス', '価格': 2980},
            {'名前': 'キーボード', '価格': 8900},
        ]
    })
    
    print("ループ結果:")
    print(result)


def include_demo():
    """インクルード機能のデモ"""
    print("=== インクルード機能 ===")
    
    engine = SimpleTemplateEngine()
    
    # インクルード用テンプレート登録
    engine.register_include('ヘッダー', """
=== {{ サイト名 }} ===
ナビゲーション: ホーム | 商品 | お問い合わせ
""")
    
    engine.register_include('フッター', """
---
© {{ 年 }} {{ サイト名 }}. All rights reserved.
お問い合わせ: {{ 連絡先 }}
---
""")
    
    # メインテンプレート
    main_template = """
{{ include "ヘッダー" }}

メインコンテンツ:
こんにちは、{{ ユーザー名 }}さん！
今日の特別オファーをご覧ください。

{{ include "フッター" }}
"""
    
    engine.set_global_variable('サイト名', 'Eコマースサイト')
    engine.set_global_variable('年', 2024)
    engine.set_global_variable('連絡先', 'support@example.com')
    
    result = engine.render(main_template, {
        'ユーザー名': 'ボブ'
    })
    
    print("インクルード結果:")
    print(result)


def layout_demo():
    """レイアウト機能のデモ"""
    print("=== レイアウト機能 ===")
    
    engine = SimpleTemplateEngine()
    
    # レイアウトテンプレート
    layout = """
<!DOCTYPE html>
<html>
<head>
    <title>{{ タイトル }} - {{ サイト名 }}</title>
</head>
<body>
    <header>
        <h1>{{ サイト名 }}</h1>
        <nav>ホーム | {{ セクション }}</nav>
    </header>
    
    <main>
{{ content }}
    </main>
    
    <footer>
        <p>© {{ 年 }} {{ サイト名 }}</p>
    </footer>
</body>
</html>
"""
    
    # コンテンツ
    content = """
        <h2>{{ ページタイトル }}</h2>
        <p>{{ メッセージ }}</p>
        
        <ul>
        {% for アイテム in リスト %}
            <li>{{ アイテム }}</li>
        {% endfor %}
        </ul>
"""
    
    engine.set_global_variable('サイト名', 'ブログサイト')
    engine.set_global_variable('年', 2024)
    
    result = engine.render_with_layout(layout, content, {
        'タイトル': 'ホームページ',
        'セクション': 'ブログ',
        'ページタイトル': '最新記事',
        'メッセージ': 'EphemeralDBの活用例をご紹介します。',
        'リスト': ['スコープ管理', 'テンプレート処理', '設定管理']
    })
    
    print("レイアウト結果:")
    print(result)


def nested_scopes_demo():
    """ネストしたスコープでの変数管理"""
    print("=== ネストしたスコープでの変数管理 ===")
    
    engine = SimpleTemplateEngine()
    
    # 複雑なネスト構造のテンプレート
    engine.register_include('商品カード', """
    <div class="商品カード">
        <h3>{{ 商品.名前 }}</h3>
        <p>価格: ¥{{ 商品.価格 }}</p>
        {% if 割引率 %}
        <p class="割引">{{ 割引率 }}%オフ！</p>
        {% endif %}
        <p>カテゴリ: {{ カテゴリ名 }}</p>
    </div>
""")
    
    main_template = """
<div class="ショップ">
    <h1>{{ ショップ名 }}</h1>
    
    {% for カテゴリ in カテゴリリスト %}
    <section class="カテゴリ">
        <h2>{{ カテゴリ.名前 }}</h2>
        
        {% for 商品 in カテゴリ.商品 %}
            {{ include "商品カード" }}
        {% endfor %}
    </section>
    {% endfor %}
</div>
"""
    
    # グローバル変数（すべてのスコープからアクセス可能）
    engine.set_global_variable('ショップ名', 'テックストア')
    engine.set_global_variable('割引率', 10)
    
    result = engine.render(main_template, {
        'カテゴリリスト': [
            {
                '名前': 'コンピューター',
                '商品': [
                    {'名前': 'ノートPC', '価格': 89800},
                    {'名前': 'デスクトップPC', '価格': 129800},
                ]
            },
            {
                '名前': '周辺機器', 
                '商品': [
                    {'名前': 'マウス', '価格': 2980},
                    {'名前': 'キーボード', '価格': 8900},
                ]
            }
        ]
    })
    
    print("ネストしたスコープ結果:")
    print(result)


def template_inheritance_demo():
    """テンプレート継承のシミュレーション"""
    print("=== テンプレート継承シミュレーション ===")
    
    class InheritanceEngine(SimpleTemplateEngine):
        """継承機能付きテンプレートエンジン"""
        
        def render_with_inheritance(self, child_template: str, parent_template: str, variables: Optional[Dict[str, Any]] = None) -> str:
            """親テンプレートを継承してレンダリング"""
            with self.db.scope('継承'):
                # 子テンプレートの変数を設定
                if variables:
                    for key, value in variables.items():
                        self.db.set(key, value)
                
                # 子テンプレートで定義されたブロックを抽出
                blocks = self._extract_blocks(child_template)
                
                # ブロックを変数として設定
                for block_name, block_content in blocks.items():
                    with self.db.scope(f'ブロック_{block_name}'):
                        processed_content = self._process_template(block_content)
                        self.db.set(f'ブロック_{block_name}', processed_content)
                
                # 親テンプレートをレンダリング
                return self._process_template(parent_template)
        
        def _extract_blocks(self, template: str) -> Dict[str, str]:
            """ブロックを抽出"""
            block_pattern = r'\{%\s*block\s+([a-zA-Z_][a-zA-Z0-9_]*)\s*%\}(.*?)\{%\s*endblock\s*%\}'
            blocks = {}
            
            for match in re.finditer(block_pattern, template, re.DOTALL):
                block_name = match.group(1)
                block_content = match.group(2).strip()
                blocks[block_name] = block_content
            
            return blocks
    
    engine = InheritanceEngine()
    
    # 基本レイアウト（親テンプレート）
    base_layout = """
<!DOCTYPE html>
<html>
<head>
    <title>{{ ブロック_title }}</title>
    <style>{{ ブロック_styles }}</style>
</head>
<body>
    <header>
        <h1>{{ サイト名 }}</h1>
    </header>
    
    <main>
{{ ブロック_content }}
    </main>
    
    <footer>
        {{ ブロック_footer }}
    </footer>
    
    <script>{{ ブロック_scripts }}</script>
</body>
</html>
"""
    
    # 子テンプレート
    child_template = """
{% block title %}{{ ページタイトル }} - ブログ{% endblock %}

{% block styles %}
body { font-family: Arial; }
.記事 { margin: 20px; }
{% endblock %}

{% block content %}
<article class="記事">
    <h2>{{ 記事タイトル }}</h2>
    <p>投稿者: {{ 投稿者 }}</p>
    <div>{{ 記事内容 }}</div>
</article>
{% endblock %}

{% block footer %}
<p>© 2024 {{ サイト名 }} - すべての権利を保有</p>
{% endblock %}

{% block scripts %}
console.log('ページが読み込まれました');
{% endblock %}
"""
    
    engine.set_global_variable('サイト名', 'テックブログ')
    
    result = engine.render_with_inheritance(child_template, base_layout, {
        'ページタイトル': 'EphemeralDB入門',
        '記事タイトル': 'EphemeralDBの基本的な使い方',
        '投稿者': '田中太郎',
        '記事内容': 'EphemeralDBは軽量で使いやすいメモリストレージです...'
    })
    
    print("テンプレート継承結果:")
    print(result)


if __name__ == "__main__":
    basic_template_demo()
    print()
    
    filter_demo()
    print()
    
    conditional_demo()
    print()
    
    loop_demo()
    print()
    
    include_demo()
    print()
    
    layout_demo()
    print()
    
    nested_scopes_demo()
    print()
    
    template_inheritance_demo()
    
    print("\n=== まとめ ===")
    print("EphemeralDBを使用したテンプレートエンジンにより、")
    print("変数スコープを適切に管理しながら複雑なテンプレート処理が可能です。")
    print("インクルード、レイアウト、継承などの高度な機能も実現できます。")


#!/usr/bin/env python3
"""
02. スコープ管理 - スコープのpush/pop操作を学ぶ
"""

from ephemeraldb import EphemeralDB


def basic_scope_operations():
    """基本的なスコープ操作"""
    print("=== 基本スコープ操作 ===")
    
    db = EphemeralDB()
    
    # 初期状態
    print(f"初期スコープ数: {db.scope_count()}")
    print(f"現在のスコープ: {db.current_scope()}")
    
    # グローバルデータを設定
    db.set('グローバル設定', 'プロダクション')
    db.set('アプリ名', 'マイアプリ')
    
    # 新しいスコープをプッシュ
    db.push_scope('ユーザーセッション')
    print(f"プッシュ後のスコープ数: {db.scope_count()}")
    print(f"現在のスコープ: {db.current_scope()}")
    
    # ローカルデータを設定
    db.set('ユーザーID', 123)
    db.set('セッショントークン', 'abc123')
    
    # 両方のスコープからデータにアクセス
    print(f"グローバル設定: {db.get('グローバル設定')}")  # 親スコープから
    print(f"ユーザーID: {db.get('ユーザーID')}")  # 現在のスコープから
    
    # スコープをポップ
    popped_data = db.pop_scope()
    print(f"ポップ後のスコープ数: {db.scope_count()}")
    print(f"ポップされたデータ: {popped_data}")
    
    # ローカルデータにアクセスできないことを確認
    print(f"ユーザーID（ポップ後）: {db.get('ユーザーID')}")  # None
    print(f"グローバル設定（ポップ後）: {db.get('グローバル設定')}")  # まだ利用可能
    
    print()


def nested_scopes():
    """ネストしたスコープの操作"""
    print("=== ネストしたスコープ ===")
    
    db = EphemeralDB()
    
    # レベル0（ルート）
    db.set('レベル0', 'ルート')
    
    # レベル1
    db.push_scope('レベル1')
    db.set('レベル1変数', 'レベル1の値')
    
    # レベル2
    db.push_scope('レベル2')
    db.set('レベル2変数', 'レベル2の値')
    
    # レベル3
    db.push_scope('レベル3')
    db.set('レベル3変数', 'レベル3の値')
    
    print(f"深いネストでのスコープ数: {db.scope_count()}")
    
    # 全レベルからデータにアクセス
    print(f"レベル0: {db.get('レベル0')}")
    print(f"レベル1変数: {db.get('レベル1変数')}")
    print(f"レベル2変数: {db.get('レベル2変数')}")
    print(f"レベル3変数: {db.get('レベル3変数')}")
    
    # 順番にスコープをポップ
    db.pop_scope()  # レベル3を削除
    print(f"レベル3削除後: {db.get('レベル3変数')}")  # None
    print(f"レベル2はまだ存在: {db.get('レベル2変数')}")  # まだある
    
    db.pop_scope()  # レベル2を削除
    print(f"レベル2削除後: {db.get('レベル2変数')}")  # None
    print(f"レベル1はまだ存在: {db.get('レベル1変数')}")  # まだある
    
    db.pop_scope()  # レベル1を削除
    print(f"レベル1削除後: {db.get('レベル1変数')}")  # None
    print(f"ルートはまだ存在: {db.get('レベル0')}")  # まだある
    
    print()


def scope_shadowing():
    """スコープのシャドウイング（同名変数の上書き）"""
    print("=== スコープのシャドウイング ===")
    
    db = EphemeralDB()
    
    # 共通キー名で異なる値を設定
    db.set('共通キー', '親の値')
    db.set('親専用', '親でのみ定義')
    
    print(f"プッシュ前の共通キー: {db.get('共通キー')}")
    
    # 子スコープで同じキーをシャドウ
    db.push_scope('子スコープ')
    db.set('共通キー', '子の値')  # 親の値をシャドウ
    db.set('子専用', '子でのみ定義')
    
    print(f"子スコープでの共通キー: {db.get('共通キー')}")  # 子の値
    print(f"子スコープでの親専用: {db.get('親専用')}")  # 親から継承
    print(f"子スコープでの子専用: {db.get('子専用')}")  # 子で定義
    
    # 子スコープをポップ
    db.pop_scope()
    
    print(f"ポップ後の共通キー: {db.get('共通キー')}")  # 親の値に戻る
    print(f"ポップ後の子専用: {db.get('子専用')}")  # None
    
    print()


def scope_isolation():
    """スコープの分離を確認"""
    print("=== スコープの分離 ===")
    
    db = EphemeralDB()
    
    db.set('グローバル', 'グローバル値')
    
    # 第1の子スコープ
    db.push_scope('子1')
    db.set('ローカル', '子1の値')
    print(f"子1でのローカル: {db.get('ローカル')}")
    db.pop_scope()
    
    # 第2の子スコープ
    db.push_scope('子2')
    db.set('ローカル', '子2の値')  # 同じキー名だが異なるスコープ
    print(f"子2でのローカル: {db.get('ローカル')}")
    print(f"子2でのグローバル: {db.get('グローバル')}")  # 親からアクセス可能
    db.pop_scope()
    
    # ルートスコープに戻る
    print(f"ルートでのローカル: {db.get('ローカル')}")  # None
    print(f"ルートでのグローバル: {db.get('グローバル')}")  # まだ存在
    
    print()


def scope_inspection():
    """スコープの検査機能"""
    print("=== スコープの検査 ===")
    
    db = EphemeralDB()
    
    # 複数のスコープを作成
    scope_names = ['認証', 'データ処理', '結果生成']
    
    for name in scope_names:
        db.push_scope(name)
        db.set(f'{name}_データ', f'{name}で生成されたデータ')
        
        print(f"現在のスコープ: {db.current_scope()}")
        print(f"スコープ数: {db.scope_count()}")
        print(f"現在のキー: {db.keys()}")
        print()
    
    # 逆順でスコープをポップ
    while db.scope_count() > 1:
        current = db.current_scope()
        data = db.pop_scope()
        print(f"'{current}'スコープをポップ: {data}")
    
    print(f"最終スコープ数: {db.scope_count()}")
    
    print()


if __name__ == "__main__":
    basic_scope_operations()
    nested_scopes()
    scope_shadowing()
    scope_isolation()
    scope_inspection()
    
    print("=== まとめ ===")
    print("スコープ管理により、変数の生存期間と可視性を制御できます。")
    print("子スコープは親スコープの変数にアクセスでき、")
    print("同名変数はシャドウイング（上書き）されます。")


#!/usr/bin/env python3
"""
05. 設定パーサー - 階層的な設定管理システム
"""

from ephemeraldb import EphemeralDB
import json
from typing import Dict, Any, List, Optional


class ConfigurationParser:
    """EphemeralDBを使用した階層的設定パーサー"""
    
    def __init__(self):
        self.db = EphemeralDB()
        self.config_stack = []
    
    def load_base_config(self, config: Dict[str, Any]):
        """ベース設定を読み込み"""
        print("=== ベース設定読み込み ===")
        for key, value in config.items():
            self.db.set(key, value)
            print(f"ベース設定: {key} = {value}")
    
    def load_environment_config(self, env_name: str, config: Dict[str, Any]):
        """環境固有の設定を読み込み"""
        with self.db.scope(f'env_{env_name}'):
            print(f"\n=== {env_name}環境設定読み込み ===")
            for key, value in config.items():
                self.db.set(key, value)
                print(f"{env_name}設定: {key} = {value}")
            
            # 現在の有効設定を表示
            self.show_effective_config(env_name)
    
    def load_user_config(self, env_name: str, user_name: str, config: Dict[str, Any]):
        """ユーザー固有の設定を読み込み"""
        with self.db.scope(f'env_{env_name}'):
            with self.db.scope(f'user_{user_name}'):
                print(f"\n=== {env_name}環境の{user_name}ユーザー設定 ===")
                for key, value in config.items():
                    self.db.set(key, value)
                    print(f"{user_name}設定: {key} = {value}")
                
                # 現在の有効設定を表示
                self.show_effective_config(f"{env_name}/{user_name}")
    
    def show_effective_config(self, context: str):
        """現在有効な設定を表示"""
        print(f"\n--- {context}での有効設定 ---")
        
        # 重要な設定項目をチェック
        important_keys = [
            'データベース.ホスト', 'データベース.ポート', 'データベース.名前',
            'API.ベースURL', 'API.タイムアウト', 'API.リトライ回数',
            'ログ.レベル', 'ログ.ファイル',
            'セキュリティ.暗号化', 'セキュリティ.認証',
            'キャッシュ.有効', 'キャッシュ.TTL'
        ]
        
        for key in important_keys:
            value = self.db.get(key)
            if value is not None:
                print(f"  {key}: {value}")
    
    def get_config_value(self, key: str, default: Any = None) -> Any:
        """設定値を取得（現在のスコープから）"""
        return self.db.get(key, default)
    
    def export_config(self, include_hierarchy: bool = True) -> Dict[str, Any]:
        """現在の設定をエクスポート"""
        return self.db.to_dict(include_hierarchy=include_hierarchy)


def basic_configuration_demo():
    """基本的な設定管理のデモ"""
    print("=== 基本設定管理デモ ===")
    
    parser = ConfigurationParser()
    
    # ベース設定
    base_config = {
        'データベース.ホスト': 'localhost',
        'データベース.ポート': 5432,
        'データベース.名前': 'myapp',
        'API.ベースURL': 'https://api.example.com',
        'API.タイムアウト': 30,
        'API.リトライ回数': 3,
        'ログ.レベル': 'INFO',
        'ログ.ファイル': '/var/log/app.log',
        'セキュリティ.暗号化': True,
        'セキュリティ.認証': 'OAuth2',
        'キャッシュ.有効': True,
        'キャッシュ.TTL': 3600
    }
    
    parser.load_base_config(base_config)
    
    # 開発環境設定
    dev_config = {
        'データベース.ホスト': 'dev-db.local',
        'データベース.名前': 'myapp_dev',
        'API.ベースURL': 'http://localhost:8000',
        'ログ.レベル': 'DEBUG',
        'ログ.ファイル': './dev.log',
        'セキュリティ.暗号化': False,  # 開発環境では無効
    }
    
    parser.load_environment_config('開発', dev_config)
    
    # 本番環境設定
    prod_config = {
        'データベース.ホスト': 'prod-db.cluster.local',
        'データベース.名前': 'myapp_production',
        'API.ベースURL': 'https://api.production.com',
        'API.タイムアウト': 60,  # 本番環境では長めのタイムアウト
        'ログ.レベル': 'WARNING',
        'ログ.ファイル': '/var/log/production/app.log',
        'セキュリティ.認証': 'JWT',  # 本番環境ではJWT
        'キャッシュ.TTL': 7200,  # 本番環境では長めのTTL
    }
    
    parser.load_environment_config('本番', prod_config)
    print()


def user_specific_config_demo():
    """ユーザー固有設定のデモ"""
    print("=== ユーザー固有設定デモ ===")
    
    parser = ConfigurationParser()
    
    # ベース設定
    base_config = {
        'UI.テーマ': '標準',
        'UI.言語': '英語',
        'UI.フォントサイズ': 14,
        '通知.メール': True,
        '通知.プッシュ': True,
        '通知.頻度': 'リアルタイム',
        'プライバシー.分析': True,
        'プライバシー.Cookie': True,
    }
    
    parser.load_base_config(base_config)
    
    # 開発環境でのユーザー設定
    alice_config = {
        'UI.テーマ': 'ダーク',
        'UI.言語': '日本語',
        'UI.フォントサイズ': 16,
        '通知.メール': False,  # アリスはメール通知を無効
        'プライバシー.分析': False,  # プライバシー重視
    }
    
    parser.load_user_config('開発', 'alice', alice_config)
    
    bob_config = {
        'UI.テーマ': 'ライト',
        'UI.フォントサイズ': 12,
        '通知.頻度': '1時間毎',  # ボブは通知頻度を下げる
    }
    
    parser.load_user_config('開発', 'bob', bob_config)
    print()


def nested_config_scopes_demo():
    """ネストした設定スコープのデモ"""
    print("=== ネストした設定スコープデモ ===")
    
    parser = ConfigurationParser()
    
    # グローバル設定
    global_config = {
        'グローバル.会社名': 'テック株式会社',
        'グローバル.バージョン': '1.0.0',
        'グローバル.サポートメール': 'support@tech.co.jp',
        'デフォルト.タイムアウト': 30,
        'デフォルト.リトライ': 3,
        'デフォルト.暗号化': True,
    }
    
    parser.load_base_config(global_config)
    
    # 地域設定（日本）
    with parser.db.scope('地域_日本'):
        region_config = {
            '地域.タイムゾーン': 'Asia/Tokyo',
            '地域.通貨': 'JPY',
            '地域.言語': '日本語',
            '地域.サポート時間': '9:00-18:00 JST',
            'デフォルト.タイムアウト': 45,  # 日本では少し長めに
        }
        
        for key, value in region_config.items():
            parser.db.set(key, value)
        
        print("=== 日本地域設定 ===")
        parser.show_effective_config('日本地域')
        
        # 都市設定（東京）
        with parser.db.scope('都市_東京'):
            city_config = {
                '都市.名前': '東京',
                '都市.データセンター': 'tokyo-1',
                '都市.CDN': 'cloudfront-ap-northeast-1',
                'デフォルト.タイムアウト': 20,  # 東京は高速回線なので短く
            }
            
            for key, value in city_config.items():
                parser.db.set(key, value)
            
            print("\n=== 東京都市設定 ===")
            parser.show_effective_config('日本地域/東京')
            
            # 特定サービス設定
            with parser.db.scope('サービス_支払い'):
                payment_config = {
                    'サービス.名前': '支払いサービス',
                    'サービス.プロバイダー': 'Stripe Japan',
                    'サービス.通貨': 'JPY',
                    'デフォルト.タイムアウト': 60,  # 支払いは慎重に
                    'セキュリティ.レベル': '最高',
                }
                
                for key, value in payment_config.items():
                    parser.db.set(key, value)
                
                print("\n=== 支払いサービス設定 ===")
                parser.show_effective_config('日本地域/東京/支払いサービス')
    
    print()


def configuration_inheritance_demo():
    """設定継承のデモ"""
    print("=== 設定継承デモ ===")
    
    class InheritanceDemo:
        def __init__(self):
            self.db = EphemeralDB()
        
        def test_inheritance_chain(self):
            """継承チェーンのテスト"""
            
            # レベル1: 企業全体の設定
            self.db.set('セキュリティ.ポリシー', '厳格')
            self.db.set('バックアップ.頻度', '毎日')
            self.db.set('ログ.保持期間', 365)
            self.db.set('タイムアウト.デフォルト', 30)
            
            print("企業レベル設定:")
            self._show_current_config(['セキュリティ.ポリシー', 'バックアップ.頻度', 'ログ.保持期間', 'タイムアウト.デフォルト'])
            
            # レベル2: 部門設定
            with self.db.scope('部門_開発'):
                self.db.set('ログ.レベル', 'DEBUG')
                self.db.set('ログ.保持期間', 90)  # 開発部門は短く
                self.db.set('タイムアウト.デフォルト', 60)  # 開発環境は長く
                
                print("\n開発部門設定:")
                self._show_current_config(['セキュリティ.ポリシー', 'バックアップ.頻度', 'ログ.レベル', 'ログ.保持期間', 'タイムアウト.デフォルト'])
                
                # レベル3: チーム設定
                with self.db.scope('チーム_フロントエンド'):
                    self.db.set('ツール.エディタ', 'VSCode')
                    self.db.set('ツール.ブラウザ', 'Chrome')
                    self.db.set('タイムアウト.デフォルト', 10)  # フロントエンドは短く
                    
                    print("\nフロントエンドチーム設定:")
                    self._show_current_config([
                        'セキュリティ.ポリシー', 'バックアップ.頻度', 'ログ.レベル', 
                        'ログ.保持期間', 'タイムアウト.デフォルト', 'ツール.エディタ', 'ツール.ブラウザ'
                    ])
                    
                    # レベル4: 個人設定
                    with self.db.scope('開発者_田中'):
                        self.db.set('個人.テーマ', 'ダーク')
                        self.db.set('個人.フォント', 'Fira Code')
                        self.db.set('ツール.エディタ', 'Neovim')  # 個人的な好み
                        
                        print("\n田中開発者の個人設定:")
                        self._show_current_config([
                            'セキュリティ.ポリシー', 'バックアップ.頻度', 'ログ.レベル', 
                            'ログ.保持期間', 'タイムアウト.デフォルト', 'ツール.エディタ', 
                            'ツール.ブラウザ', '個人.テーマ', '個人.フォント'
                        ])
        
        def _show_current_config(self, keys: List[str]):
            """現在の設定を表示"""
            for key in keys:
                value = self.db.get(key)
                print(f"  {key}: {value}")
    
    demo = InheritanceDemo()
    demo.test_inheritance_chain()
    print()


def configuration_templates_demo():
    """設定テンプレートのデモ"""
    print("=== 設定テンプレートデモ ===")
    
    class ConfigTemplate:
        def __init__(self, name: str):
            self.name = name
            self.db = EphemeralDB()
        
        def apply_template(self, template_name: str, template_config: Dict[str, Any]):
            """テンプレートを適用"""
            with self.db.scope(f'テンプレート_{template_name}'):
                print(f"=== {template_name}テンプレート適用 ===")
                for key, value in template_config.items():
                    self.db.set(key, value)
                    print(f"  {key}: {value}")
                
                return self.get_effective_config()
        
        def get_effective_config(self) -> Dict[str, Any]:
            """有効な設定を取得"""
            return self.db.to_dict(include_hierarchy=True)
    
    # Webアプリケーション用テンプレート
    web_template = {
        'サーバー.ポート': 8080,
        'サーバー.ワーカー数': 4,
        'データベース.接続プール': 10,
        'セッション.タイムアウト': 1800,
        '静的ファイル.キャッシュ': True,
        'ログ.レベル': 'INFO',
        'セキュリティ.HTTPS': True,
        'セキュリティ.CSRF保護': True,
    }
    
    # API用テンプレート
    api_template = {
        'サーバー.ポート': 3000,
        'サーバー.ワーカー数': 8,
        'データベース.接続プール': 20,
        'API.レート制限': 1000,
        'API.CORS': True,
        'ログ.レベル': 'DEBUG',
        'セキュリティ.JWT': True,
        'セキュリティ.API Key': True,
    }
    
    # マイクロサービス用テンプレート
    microservice_template = {
        'サーバー.ポート': 9000,
        'サーバー.ワーカー数': 2,
        'データベース.接続プール': 5,
        'サービス.名前': 'unknown',
        'サービス.バージョン': '1.0.0',
        'ヘルスチェック.間隔': 30,
        'メトリクス.有効': True,
        'トレーシング.有効': True,
    }
    
    # 各テンプレートをテスト
    config = ConfigTemplate('テストアプリ')
    
    # ベース設定
    base_config = {
        'アプリ.名前': 'サンプルアプリ',
        'アプリ.バージョン': '2.0.0',
        'ログ.ファイル': '/var/log/app.log',
    }
    
    for key, value in base_config.items():
        config.db.set(key, value)
    
    print("ベース設定:")
    print(json.dumps(config.get_effective_config(), indent=2, ensure_ascii=False))
    
    # Webテンプレート適用
    web_config = config.apply_template('Web', web_template)
    print("\nWebテンプレート適用後:")
    print(json.dumps(web_config, indent=2, ensure_ascii=False))
    
    # APIテンプレート適用（新しいスコープ）
    api_config = config.apply_template('API', api_template)
    print("\nAPIテンプレート適用後:")
    print(json.dumps(api_config, indent=2, ensure_ascii=False))
    
    print()


if __name__ == "__main__":
    basic_configuration_demo()
    user_specific_config_demo()
    nested_config_scopes_demo()
    configuration_inheritance_demo()
    configuration_templates_demo()
    
    print("=== まとめ ===")
    print("EphemeralDBを使用することで、複雑な階層構造を持つ")
    print("設定管理システムを簡単に実装できます。")
    print("環境、ユーザー、地域などの様々な条件に応じた")
    print("設定の継承と上書きが自然に表現できます。")


"""
Pytest configuration and fixtures for EphemeralDB tests
"""

import pytest
from ephemeraldb import EphemeralDB


@pytest.fixture
def db():
    """Provide a fresh EphemeralDB instance for each test"""
    return EphemeralDB()


@pytest.fixture
def populated_db():
    """Provide an EphemeralDB instance with some test data"""
    db = EphemeralDB()
    
    # Add some test data
    db.set('string_key', 'string_value')
    db.set('int_key', 42)
    db.set('list_key', [1, 2, 3])
    db.set('dict_key', {'nested': 'value'})
    db.set('nested.key1', 'nested_value1')
    db.set('nested.key2', 'nested_value2')
    db.set('deep.nested.key', 'deep_value')
    
    return db


@pytest.fixture
def multi_scope_db():
    """Provide an EphemeralDB instance with multiple scopes set up"""
    db = EphemeralDB()
    
    # Root scope data
    db.set('global_var', 'global_value')
    db.set('shared_key', 'root_value')
    
    # Level 1 scope
    db.push_scope('level1')
    db.set('level1_var', 'level1_value')
    db.set('shared_key', 'level1_value')  # Shadow root value
    
    # Level 2 scope
    db.push_scope('level2')
    db.set('level2_var', 'level2_value')
    
    return db


"""
Test the examples from README to ensure they work as documented
"""

import pytest
from ephemeraldb import EphemeralDB


class TestREADMEExamples:
    """Test all examples from the README"""
    
    def test_basic_usage_example(self):
        """Test the basic usage example from README"""
        # Create a new database instance
        db = EphemeralDB()

        # Basic usage
        db.set('name', 'Alice')
        db.set('user.age', 30)
        assert db.get('name') == 'Alice'
        assert db.get('user.age') == 30
    
    def test_scoped_usage_example(self):
        """Test the scoped usage example from README"""
        db = EphemeralDB()
        
        # Scoped usage
        db.set('global_var', 'I am global')

        db.push_scope('scope1')
        db.set('local_var', 'I am local')
        assert db.get('global_var') == 'I am global'  # accessible from parent
        assert db.get('local_var') == 'I am local'

        db.pop_scope()
        assert db.get('local_var') is None  # scope popped
    
    def test_context_manager_example(self):
        """Test the context manager example from README"""
        db = EphemeralDB()
        
        # Context manager usage
        with db.scope('temp_scope'):
            db.set('temp_data', 'temporary')
            assert db.get('temp_data') == 'temporary'
        
        # temp_data is automatically cleaned up
        assert db.get('temp_data') is None
    
    def test_dsl_interpreter_usecase(self):
        """Test a DSL interpreter use case"""
        db = EphemeralDB()
        
        # Simulate DSL interpreter with variable scoping
        def execute_block(variables, statements):
            """Simulate executing a block of statements with local variables"""
            with db.scope('block_scope'):
                # Import block variables
                for var, value in variables.items():
                    db.set(var, value)
                
                results = []
                for statement in statements:
                    if statement.startswith('SET '):
                        _, var, value = statement.split(' ', 2)
                        db.set(var, value)
                    elif statement.startswith('GET '):
                        _, var = statement.split(' ', 1)
                        results.append(db.get(var))
                
                return results
        
        # Set global variables
        db.set('global_constant', 42)
        
        # Execute nested blocks
        results1 = execute_block(
            {'local_var': 'block1'},
            ['SET temp_var hello', 'GET local_var', 'GET temp_var', 'GET global_constant']
        )
        
        assert results1 == ['block1', 'hello', 42]
        
        # Variables should be cleaned up
        assert db.get('local_var') is None
        assert db.get('temp_var') is None
        assert db.get('global_constant') == 42
    
    def test_configuration_parser_usecase(self):
        """Test a configuration parser use case"""
        db = EphemeralDB()
        
        def parse_config_section(section_name, config_data):
            """Parse a configuration section with inheritance"""
            with db.scope(f'config_{section_name}'):
                for key, value in config_data.items():
                    db.set(f'{section_name}.{key}', value)
                
                # Return resolved configuration
                return {
                    'section': section_name,
                    'resolved_config': db.to_dict(include_hierarchy=True)
                }
        
        # Set global defaults
        db.set('defaults.timeout', 30)
        db.set('defaults.retries', 3)
        
        # Parse environment-specific configs
        dev_config = parse_config_section('development', {
            'host': 'localhost',
            'port': 8000,
            'debug': True
        })
        
        prod_config = parse_config_section('production', {
            'host': 'prod.example.com',
            'port': 443,
            'timeout': 60  # Override default
        })
        
        # Verify inheritance works
        assert dev_config['resolved_config']['defaults.timeout'] == 30
        assert dev_config['resolved_config']['development.debug'] is True
        
        assert prod_config['resolved_config']['defaults.retries'] == 3
        assert prod_config['resolved_config']['production.timeout'] == 60
    
    def test_nested_transaction_usecase(self):
        """Test nested transaction processing use case"""
        db = EphemeralDB()
        
        class Transaction:
            def __init__(self, db, name):
                self.db = db
                self.name = name
                self.changes = []
            
            def __enter__(self):
                self.db.push_scope(f'transaction_{self.name}')
                return self
            
            def __exit__(self, exc_type, exc_val, exc_tb):
                if exc_type is None:
                    # Transaction succeeded, record changes
                    self.changes = list(self.db.items(include_nested=False))
                else:
                    # Transaction failed, changes will be rolled back automatically
                    pass
                self.db.pop_scope()
        
        # Set initial state
        db.set('account.balance', 1000)
        db.set('account.transactions', [])
        
        # Nested transactions
        try:
            with Transaction(db, 'outer') as outer_tx:
                db.set('pending.amount', 200)
                
                with Transaction(db, 'inner') as inner_tx:
                    current_balance = db.get('account.balance')
                    pending_amount = db.get('pending.amount')
                    
                    if current_balance >= pending_amount:
                        db.set('account.balance', current_balance - pending_amount)
                        db.set('temp.new_balance', current_balance - pending_amount)
                    else:
                        raise ValueError("Insufficient funds")
                
                # Inner transaction succeeded
                new_balance = db.get('temp.new_balance')
                if new_balance is not None:
                    db.set('account.balance', new_balance)
        
        except ValueError:
            # Transaction rolled back
            pass
        
        # Verify final state
        assert db.get('account.balance') == 1000  # Transaction was rolled back due to scope cleanup
        assert db.get('pending.amount') is None  # Cleaned up
        assert db.get('temp.new_balance') is None  # Cleaned up


class TestAdvancedUseCases:
    """Test advanced use cases and patterns"""
    
    def test_template_engine_context(self):
        """Test template engine variable context management"""
        db = EphemeralDB()
        
        def render_template(template, context):
            """Simulate template rendering with variable context"""
            with db.scope('template_context'):
                # Set template variables
                for key, value in context.items():
                    db.set(key, value)
                
                # Simple template variable substitution
                result = template
                # Get all keys from current scope and parent scopes
                all_keys = set()
                for scope in reversed(db._scopes):
                    for key in scope.keys():
                        if '.' not in key:  # Only top-level variables
                            all_keys.add(key)
                
                for key in all_keys:
                    value = db.get(key)
                    if value is not None:
                        result = result.replace(f'{{{key}}}', str(value))
                
                return result
        
        # Set global template variables
        db.set('site_name', 'My Website')
        db.set('copyright_year', 2024)
        
        # Render templates with local context
        template1 = "Welcome to {site_name}! Hello {user_name}!"
        result1 = render_template(template1, {'user_name': 'Alice'})
        assert result1 == "Welcome to My Website! Hello Alice!"
        
        template2 = "© {copyright_year} {site_name}. User: {user_name}"
        result2 = render_template(template2, {'user_name': 'Bob'})
        assert result2 == "© 2024 My Website. User: Bob"
        
        # Verify isolation
        assert db.get('user_name') is None
    
    def test_recursive_function_context(self):
        """Test recursive function with context preservation"""
        db = EphemeralDB()
        
        def recursive_processor(data, depth=0):
            """Process nested data with depth-aware context"""
            with db.scope(f'depth_{depth}'):
                db.set('current_depth', depth)
                
                # Update max_depth in root scope, not current scope
                current_max = db.get('max_depth', 0)
                if depth > current_max:
                    # Set max_depth in root scope
                    if len(db._scopes) > 1:
                        db._scopes[0]['max_depth'] = depth
                    else:
                        db.set('max_depth', depth)
                
                if isinstance(data, dict):
                    result = {}
                    for key, value in data.items():
                        db.set(f'processing_key', key)
                        result[key] = recursive_processor(value, depth + 1)
                    return result
                elif isinstance(data, list):
                    result = []
                    for i, item in enumerate(data):
                        db.set(f'processing_index', i)
                        result.append(recursive_processor(item, depth + 1))
                    return result
                else:
                    # Leaf node - add depth info
                    current_depth = db.get('current_depth')
                    return f"{data}_depth{current_depth}"
        
        # Test with nested structure
        test_data = {
            'level1': {
                'level2': {
                    'level3': 'deep_value'
                },
                'list': [1, 2, {'nested_in_list': 'value'}]
            },
            'simple': 'simple_value'
        }
        
        result = recursive_processor(test_data)
        
        # Verify processing worked correctly
        assert result['level1']['level2']['level3'] == 'deep_value_depth3'
        assert result['level1']['list'][0] == '1_depth3'  # Items in list are processed at depth 3
        assert result['level1']['list'][2]['nested_in_list'] == 'value_depth4'  # Nested in list at depth 4
        assert result['simple'] == 'simple_value_depth1'
        
        # Verify context was cleaned up
        assert db.get('current_depth') is None
        assert db.get('processing_key') is None
        assert db.get('max_depth') == 4  # This should persist as it was set in root scope (depth 4 for nested_in_list)


"""
Core implementation of EphemeralDB
"""

import threading
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Iterator, Tuple
from collections import defaultdict

from .exceptions import ScopeError, KeyError as EphemeralKeyError, ValidationError, CapacityError


class EphemeralDB:
    """
    A lightweight volatile context management store with scoped key-value storage.
    
    Features:
    - Hierarchical scoping with push/pop operations
    - Dot notation support for nested keys
    - Thread-safe operations
    - Context manager support
    - Parent scope data inheritance
    """
    
    def __init__(self, max_scopes: int = 1000, max_keys_per_scope: int = 10000):
        """
        Initialize EphemeralDB with a root scope.
        
        Args:
            max_scopes: Maximum number of scopes allowed (default: 1000)
            max_keys_per_scope: Maximum number of keys per scope (default: 10000)
        """
        if max_scopes < 1:
            raise ValidationError("max_scopes must be at least 1", "INVALID_MAX_SCOPES")
        if max_keys_per_scope < 1:
            raise ValidationError("max_keys_per_scope must be at least 1", "INVALID_MAX_KEYS")
        
        self._lock = threading.RLock()
        self._scopes: List[Dict[str, Any]] = [{}]  # Stack of scopes, root scope at index 0
        self._scope_names: List[Optional[str]] = [None]  # Names of scopes
        self._max_scopes = max_scopes
        self._max_keys_per_scope = max_keys_per_scope
    
    def _parse_key(self, key: str) -> List[str]:
        """Parse a dot-separated key into components."""
        if not isinstance(key, str):
            raise EphemeralKeyError(
                f"キーは文字列である必要があります。{type(key).__name__}型が指定されました", 
                "INVALID_KEY_TYPE",
                {"provided_type": type(key).__name__, "expected_type": "str"}
            )
        
        if not key:
            raise EphemeralKeyError("キーは空文字列にできません", "EMPTY_KEY")
        
        if len(key) > 1000:  # Reasonable limit for key length
            raise EphemeralKeyError(
                f"キーが長すぎます（最大1000文字）: {len(key)}文字", 
                "KEY_TOO_LONG",
                {"key_length": len(key), "max_length": 1000}
            )
        
        return key.split('.')
    
    def _check_capacity_limits(self, scope_index: int = -1) -> None:
        """Check if capacity limits are exceeded."""
        current_scope = self._scopes[scope_index]
        key_count = len(self.keys())
        
        if key_count >= self._max_keys_per_scope:
            raise CapacityError(
                f"スコープ内のキー数が上限を超えました: {key_count}/{self._max_keys_per_scope}",
                "MAX_KEYS_EXCEEDED",
                {"current_keys": key_count, "max_keys": self._max_keys_per_scope}
            )
    
    def _get_nested_value(self, data: Dict[str, Any], key_parts: List[str]) -> Any:
        """Get a nested value from a dictionary using key parts."""
        current = data
        for part in key_parts:
            if not isinstance(current, dict) or part not in current:
                return None
            current = current[part]
        return current
    
    def _set_nested_value(self, data: Dict[str, Any], key_parts: List[str], value: Any) -> None:
        """Set a nested value in a dictionary using key parts."""
        current = data
        for part in key_parts[:-1]:
            if part not in current:
                current[part] = {}
            elif not isinstance(current[part], dict):
                current[part] = {}
            current = current[part]
        current[key_parts[-1]] = value
    
    def _delete_nested_value(self, data: Dict[str, Any], key_parts: List[str]) -> bool:
        """Delete a nested value from a dictionary using key parts."""
        if len(key_parts) == 1:
            if key_parts[0] in data:
                del data[key_parts[0]]
                return True
            return False
        
        current = data
        for part in key_parts[:-1]:
            if not isinstance(current, dict) or part not in current:
                return False
            current = current[part]
        
        if isinstance(current, dict) and key_parts[-1] in current:
            del current[key_parts[-1]]
            return True
        return False
    
    def set(self, key: str, value: Any) -> None:
        """
        Store a value in the current scope.
        
        Args:
            key: The key to store the value under (supports dot notation)
            value: The value to store
            
        Raises:
            EphemeralKeyError: If key is invalid
            CapacityError: If capacity limits are exceeded
        """
        with self._lock:
            key_parts = self._parse_key(key)
            
            # Check capacity before adding
            self._check_capacity_limits()
            
            current_scope = self._scopes[-1]
            self._set_nested_value(current_scope, key_parts, value)
    
    def get(self, key: str, default: Any = None) -> Any:
        """
        Retrieve a value from the current scope or parent scopes.
        
        Args:
            key: The key to retrieve (supports dot notation)
            default: Default value if key is not found
            
        Returns:
            The stored value or default
        """
        with self._lock:
            key_parts = self._parse_key(key)
            
            # Search from current scope up to root scope
            for scope in reversed(self._scopes):
                value = self._get_nested_value(scope, key_parts)
                if value is not None:
                    return value
            
            return default
    
    def delete(self, key: str) -> bool:
        """
        Delete a key from the current scope.
        
        Args:
            key: The key to delete (supports dot notation)
            
        Returns:
            True if key was deleted, False if key was not found
        """
        with self._lock:
            key_parts = self._parse_key(key)
            current_scope = self._scopes[-1]
            return self._delete_nested_value(current_scope, key_parts)
    
    def exists(self, key: str) -> bool:
        """
        Check if a key exists in current scope or parent scopes.
        
        Args:
            key: The key to check (supports dot notation)
            
        Returns:
            True if key exists, False otherwise
        """
        with self._lock:
            key_parts = self._parse_key(key)
            
            # Search from current scope up to root scope
            for scope in reversed(self._scopes):
                value = self._get_nested_value(scope, key_parts)
                if value is not None:
                    return True
            
            return False
    
    def clear(self) -> None:
        """Clear all data in the current scope."""
        with self._lock:
            self._scopes[-1].clear()
    
    def push_scope(self, name: Optional[str] = None) -> None:
        """
        Create a new scope and push it onto the stack.
        
        Args:
            name: Optional name for the scope
            
        Raises:
            CapacityError: If maximum number of scopes is exceeded
            ValidationError: If scope name is invalid
        """
        with self._lock:
            if len(self._scopes) >= self._max_scopes:
                raise CapacityError(
                    f"最大スコープ数を超えました: {len(self._scopes)}/{self._max_scopes}",
                    "MAX_SCOPES_EXCEEDED",
                    {"current_scopes": len(self._scopes), "max_scopes": self._max_scopes}
                )
            
            if name is not None:
                if not isinstance(name, str):
                    raise ValidationError(
                        f"スコープ名は文字列である必要があります: {type(name).__name__}型が指定されました",
                        "INVALID_SCOPE_NAME_TYPE",
                        {"provided_type": type(name).__name__}
                    )
                
                if len(name) > 100:  # Reasonable limit
                    raise ValidationError(
                        f"スコープ名が長すぎます（最大100文字）: {len(name)}文字",
                        "SCOPE_NAME_TOO_LONG",
                        {"name_length": len(name), "max_length": 100}
                    )
            
            self._scopes.append({})
            self._scope_names.append(name)
    
    def pop_scope(self) -> Dict[str, Any]:
        """
        Remove the current scope from the stack.
        
        Returns:
            The data from the popped scope
            
        Raises:
            ScopeError: If trying to pop the root scope
        """
        with self._lock:
            if len(self._scopes) <= 1:
                raise ScopeError(
                    "ルートスコープを削除できません",
                    "CANNOT_POP_ROOT_SCOPE",
                    {"current_scope_count": len(self._scopes)}
                )
            
            self._scope_names.pop()
            return self._scopes.pop()
    
    def current_scope(self) -> Optional[str]:
        """
        Get the name of the current scope.
        
        Returns:
            The name of the current scope or None if unnamed
        """
        with self._lock:
            return self._scope_names[-1]
    
    def scope_count(self) -> int:
        """
        Get the number of active scopes.
        
        Returns:
            The number of scopes (including root)
        """
        with self._lock:
            return len(self._scopes)
    
    @contextmanager
    def scope(self, name: Optional[str] = None) -> Iterator['EphemeralDB']:
        """
        Context manager for creating temporary scopes.
        
        Args:
            name: Optional name for the scope
            
        Yields:
            The EphemeralDB instance
        """
        self.push_scope(name)
        try:
            yield self
        finally:
            self.pop_scope()
    
    def keys(self, include_nested: bool = True) -> List[str]:
        """
        Get all keys in the current scope.
        
        Args:
            include_nested: If True, include nested keys with dot notation
            
        Returns:
            List of keys
        """
        with self._lock:
            current_scope = self._scopes[-1]
            
            if not include_nested:
                return list(current_scope.keys())
            
            keys = []
            
            def extract_keys(data: Dict[str, Any], prefix: str = ""):
                for key, value in data.items():
                    full_key = f"{prefix}.{key}" if prefix else key
                    
                    if isinstance(value, dict):
                        extract_keys(value, full_key)
                    else:
                        keys.append(full_key)
            
            extract_keys(current_scope)
            return keys
    
    def items(self, include_nested: bool = True) -> List[Tuple[str, Any]]:
        """
        Get all key-value pairs in the current scope.
        
        Args:
            include_nested: If True, include nested keys with dot notation
            
        Returns:
            List of (key, value) tuples
        """
        with self._lock:
            current_scope = self._scopes[-1]
            
            if not include_nested:
                return list(current_scope.items())
            
            items = []
            
            def extract_items(data: Dict[str, Any], prefix: str = ""):
                for key, value in data.items():
                    full_key = f"{prefix}.{key}" if prefix else key
                    
                    if isinstance(value, dict):
                        extract_items(value, full_key)
                    else:
                        items.append((full_key, value))
            
            extract_items(current_scope)
            return items
    
    def to_dict(self, include_hierarchy: bool = False) -> Dict[str, Any]:
        """
        Export current scope data as a dictionary.
        
        Args:
            include_hierarchy: If True, include data from parent scopes
            
        Returns:
            Dictionary containing the data
        """
        with self._lock:
            if not include_hierarchy:
                return dict(self._scopes[-1])
            
            # Merge all scopes from root to current, flattening nested keys
            merged = {}
            
            def flatten_dict(data: Dict[str, Any], prefix: str = ""):
                for key, value in data.items():
                    full_key = f"{prefix}.{key}" if prefix else key
                    if isinstance(value, dict):
                        flatten_dict(value, full_key)
                    else:
                        merged[full_key] = value
            
            for scope in self._scopes:
                flatten_dict(scope)
            
            return merged
    
    def from_dict(self, data: Dict[str, Any]) -> None:
        """
        Import dictionary data into the current scope.
        
        Args:
            data: Dictionary to import
        """
        with self._lock:
            current_scope = self._scopes[-1]
            current_scope.update(data)
    
    def __repr__(self) -> str:
        """String representation of EphemeralDB."""
        with self._lock:
            scope_info = f"scopes={len(self._scopes)}"
            current_name = self.current_scope()
            if current_name:
                scope_info += f", current='{current_name}'"
            
            return f"EphemeralDB({scope_info})"
    
    def __len__(self) -> int:
        """Get the number of keys in the current scope."""
        with self._lock:
            return len(self.keys())
    
    def __contains__(self, key: str) -> bool:
        """Check if a key exists (supports 'in' operator)."""
        return self.exists(key)


"""
Tests for EphemeralDB core functionality
"""

import pytest
import threading
import time
from unittest.mock import patch

from ephemeraldb import EphemeralDB
from ephemeraldb.exceptions import ScopeError, KeyError as EphemeralKeyError


class TestBasicOperations:
    """Test basic CRUD operations"""
    
    def test_set_and_get(self):
        """Test basic set and get operations"""
        db = EphemeralDB()
        
        db.set('key1', 'value1')
        assert db.get('key1') == 'value1'
        
        db.set('key2', 42)
        assert db.get('key2') == 42
        
        db.set('key3', [1, 2, 3])
        assert db.get('key3') == [1, 2, 3]
    
    def test_get_default_value(self):
        """Test get with default values"""
        db = EphemeralDB()
        
        assert db.get('nonexistent') is None
        assert db.get('nonexistent', 'default') == 'default'
        assert db.get('nonexistent', 42) == 42
    
    def test_delete(self):
        """Test delete operations"""
        db = EphemeralDB()
        
        db.set('key1', 'value1')
        assert db.exists('key1')
        
        result = db.delete('key1')
        assert result is True
        assert not db.exists('key1')
        
        # Delete non-existent key
        result = db.delete('nonexistent')
        assert result is False
    
    def test_exists(self):
        """Test exists operations"""
        db = EphemeralDB()
        
        assert not db.exists('key1')
        
        db.set('key1', 'value1')
        assert db.exists('key1')
        
        db.delete('key1')
        assert not db.exists('key1')
    
    def test_clear(self):
        """Test clear operation"""
        db = EphemeralDB()
        
        db.set('key1', 'value1')
        db.set('key2', 'value2')
        
        assert db.exists('key1')
        assert db.exists('key2')
        
        db.clear()
        
        assert not db.exists('key1')
        assert not db.exists('key2')


class TestNestedKeys:
    """Test dot notation and nested key functionality"""
    
    def test_nested_key_operations(self):
        """Test nested key operations with dot notation"""
        db = EphemeralDB()
        
        db.set('user.name', 'Alice')
        db.set('user.profile.age', 30)
        db.set('user.profile.email', 'alice@example.com')
        
        assert db.get('user.name') == 'Alice'
        assert db.get('user.profile.age') == 30
        assert db.get('user.profile.email') == 'alice@example.com'
    
    def test_nested_key_overwrite(self):
        """Test overwriting nested structures"""
        db = EphemeralDB()
        
        db.set('config.database.host', 'localhost')
        db.set('config.database.port', 5432)
        
        # Overwrite entire database config
        db.set('config.database', 'sqlite:///test.db')
        
        assert db.get('config.database') == 'sqlite:///test.db'
        assert db.get('config.database.host') is None
    
    def test_nested_key_delete(self):
        """Test deleting nested keys"""
        db = EphemeralDB()
        
        db.set('app.settings.debug', True)
        db.set('app.settings.port', 8000)
        db.set('app.name', 'MyApp')
        
        # Delete specific nested key
        result = db.delete('app.settings.debug')
        assert result is True
        assert db.get('app.settings.debug') is None
        assert db.get('app.settings.port') == 8000
        assert db.get('app.name') == 'MyApp'
    
    def test_nested_key_exists(self):
        """Test exists with nested keys"""
        db = EphemeralDB()
        
        db.set('level1.level2.level3', 'deep_value')
        
        assert db.exists('level1.level2.level3')
        assert not db.exists('level1.level2.level4')
        assert not db.exists('level1.nonexistent')


class TestScopeManagement:
    """Test scope push/pop functionality"""
    
    def test_basic_scope_operations(self):
        """Test basic scope push and pop"""
        db = EphemeralDB()
        
        # Initially has root scope
        assert db.scope_count() == 1
        assert db.current_scope() is None
        
        # Push a scope
        db.push_scope('scope1')
        assert db.scope_count() == 2
        assert db.current_scope() == 'scope1'
        
        # Pop the scope
        popped_data = db.pop_scope()
        assert db.scope_count() == 1
        assert db.current_scope() is None
        assert isinstance(popped_data, dict)
    
    def test_scope_data_isolation(self):
        """Test that scope data is properly isolated"""
        db = EphemeralDB()
        
        # Set data in root scope
        db.set('global_var', 'global_value')
        
        # Push scope and set local data
        db.push_scope('local_scope')
        db.set('local_var', 'local_value')
        
        # Both should be accessible
        assert db.get('global_var') == 'global_value'
        assert db.get('local_var') == 'local_value'
        
        # Pop scope
        db.pop_scope()
        
        # Only global should be accessible
        assert db.get('global_var') == 'global_value'
        assert db.get('local_var') is None
    
    def test_nested_scopes(self):
        """Test multiple nested scopes"""
        db = EphemeralDB()
        
        db.set('level0', 'root')
        
        db.push_scope('level1')
        db.set('level1_var', 'level1_value')
        
        db.push_scope('level2')
        db.set('level2_var', 'level2_value')
        
        # All levels should be accessible
        assert db.get('level0') == 'root'
        assert db.get('level1_var') == 'level1_value'
        assert db.get('level2_var') == 'level2_value'
        assert db.scope_count() == 3
        
        # Pop one level
        db.pop_scope()
        assert db.get('level0') == 'root'
        assert db.get('level1_var') == 'level1_value'
        assert db.get('level2_var') is None
        assert db.scope_count() == 2
        
        # Pop another level
        db.pop_scope()
        assert db.get('level0') == 'root'
        assert db.get('level1_var') is None
        assert db.scope_count() == 1
    
    def test_cannot_pop_root_scope(self):
        """Test that root scope cannot be popped"""
        db = EphemeralDB()
        
        with pytest.raises(ScopeError, match="ルートスコープを削除できません"):
            db.pop_scope()
    
    def test_scope_shadowing(self):
        """Test that child scopes can shadow parent scope values"""
        db = EphemeralDB()
        
        db.set('shared_key', 'parent_value')
        
        db.push_scope('child')
        db.set('shared_key', 'child_value')
        
        # Child value should be returned
        assert db.get('shared_key') == 'child_value'
        
        db.pop_scope()
        
        # Parent value should be restored
        assert db.get('shared_key') == 'parent_value'


class TestContextManager:
    """Test context manager functionality"""
    
    def test_context_manager_basic(self):
        """Test basic context manager usage"""
        db = EphemeralDB()
        
        db.set('global_var', 'global')
        
        with db.scope('temp_scope'):
            assert db.current_scope() == 'temp_scope'
            db.set('temp_var', 'temporary')
            assert db.get('temp_var') == 'temporary'
            assert db.get('global_var') == 'global'
        
        # After context exit
        assert db.current_scope() is None
        assert db.get('temp_var') is None
        assert db.get('global_var') == 'global'
    
    def test_context_manager_exception(self):
        """Test context manager cleanup on exception"""
        db = EphemeralDB()
        
        initial_count = db.scope_count()
        
        try:
            with db.scope('error_scope'):
                db.set('temp_data', 'will_be_lost')
                raise ValueError("Test exception")
        except ValueError:
            pass
        
        # Scope should be cleaned up even after exception
        assert db.scope_count() == initial_count
        assert db.get('temp_data') is None
    
    def test_nested_context_managers(self):
        """Test nested context managers"""
        db = EphemeralDB()
        
        db.set('level0', 'root')
        
        with db.scope('level1'):
            db.set('level1_var', 'level1')
            
            with db.scope('level2'):
                db.set('level2_var', 'level2')
                
                assert db.get('level0') == 'root'
                assert db.get('level1_var') == 'level1'
                assert db.get('level2_var') == 'level2'
                assert db.scope_count() == 3
            
            # After inner context
            assert db.get('level2_var') is None
            assert db.get('level1_var') == 'level1'
            assert db.scope_count() == 2
        
        # After outer context
        assert db.get('level1_var') is None
        assert db.scope_count() == 1


class TestUtilityMethods:
    """Test utility and introspection methods"""
    
    def test_keys_method(self):
        """Test keys() method"""
        db = EphemeralDB()
        
        db.set('simple_key', 'value')
        db.set('nested.key', 'nested_value')
        db.set('deeply.nested.key', 'deep_value')
        
        keys = db.keys()
        assert 'simple_key' in keys
        assert 'nested.key' in keys
        assert 'deeply.nested.key' in keys
        
        # Test without nested keys
        keys_flat = db.keys(include_nested=False)
        assert 'simple_key' in keys_flat
        assert 'nested' in keys_flat
        assert 'deeply' in keys_flat
        assert 'nested.key' not in keys_flat
    
    def test_items_method(self):
        """Test items() method"""
        db = EphemeralDB()
        
        db.set('key1', 'value1')
        db.set('nested.key2', 'value2')
        
        items = db.items()
        
        # Should include nested items as flattened
        items_dict = dict(items)
        assert items_dict['key1'] == 'value1'
        assert items_dict['nested.key2'] == 'value2'
    
    def test_to_dict_method(self):
        """Test to_dict() method"""
        db = EphemeralDB()
        
        db.set('root_key', 'root_value')
        
        db.push_scope('child')
        db.set('child_key', 'child_value')
        
        # Current scope only
        current_dict = db.to_dict()
        assert 'child_key' in current_dict
        assert 'root_key' not in current_dict
        
        # Include hierarchy
        full_dict = db.to_dict(include_hierarchy=True)
        assert 'child_key' in full_dict
        assert 'root_key' in full_dict
    
    def test_from_dict_method(self):
        """Test from_dict() method"""
        db = EphemeralDB()
        
        data = {
            'key1': 'value1',
            'key2': 42,
            'nested': {'inner': 'inner_value'}
        }
        
        db.from_dict(data)
        
        assert db.get('key1') == 'value1'
        assert db.get('key2') == 42
        assert db.get('nested') == {'inner': 'inner_value'}
    
    def test_len_and_contains(self):
        """Test __len__ and __contains__ methods"""
        db = EphemeralDB()
        
        assert len(db) == 0
        
        db.set('key1', 'value1')
        db.set('nested.key2', 'value2')
        
        # len() counts leaf keys only (not intermediate dictionary nodes)
        assert len(db) == 2
        assert 'key1' in db
        assert 'nested.key2' in db
        assert 'nonexistent' not in db
    
    def test_repr(self):
        """Test __repr__ method"""
        db = EphemeralDB()
        
        repr_str = repr(db)
        assert 'EphemeralDB' in repr_str
        assert 'scopes=1' in repr_str
        
        db.push_scope('named_scope')
        repr_str = repr(db)
        assert 'scopes=2' in repr_str
        assert "current='named_scope'" in repr_str


class TestErrorHandling:
    """Test error handling and edge cases"""
    
    def test_invalid_key_type(self):
        """Test handling of invalid key types"""
        db = EphemeralDB()
        
        with pytest.raises(EphemeralKeyError):
            db.set(123, 'value')
        
        with pytest.raises(EphemeralKeyError):
            db.get(None)
        
        with pytest.raises(EphemeralKeyError):
            db.delete([1, 2, 3])
    
    def test_empty_key(self):
        """Test handling of empty keys"""
        db = EphemeralDB()
        
        # Empty string should raise an error now
        with pytest.raises(EphemeralKeyError, match="キーは空文字列にできません"):
            db.set('', 'empty_key_value')
    
    def test_complex_nested_structures(self):
        """Test complex nested data structures"""
        db = EphemeralDB()
        
        complex_data = {
            'users': [
                {'name': 'Alice', 'age': 30},
                {'name': 'Bob', 'age': 25}
            ],
            'config': {
                'database': {
                    'host': 'localhost',
                    'port': 5432,
                    'credentials': {
                        'username': 'admin',
                        'password': 'secret'
                    }
                }
            }
        }
        
        db.set('app_data', complex_data)
        retrieved_data = db.get('app_data')
        
        assert retrieved_data == complex_data
        assert retrieved_data['users'][0]['name'] == 'Alice'
        assert retrieved_data['config']['database']['credentials']['username'] == 'admin'


class TestThreadSafety:
    """Test thread safety"""
    
    def test_concurrent_operations(self):
        """Test concurrent read/write operations"""
        db = EphemeralDB()
        results = []
        errors = []
        
        def worker(worker_id):
            try:
                for i in range(100):
                    key = f'worker_{worker_id}_key_{i}'
                    value = f'worker_{worker_id}_value_{i}'
                    
                    db.set(key, value)
                    retrieved = db.get(key)
                    
                    if retrieved == value:
                        results.append(f'{worker_id}_{i}_success')
                    else:
                        errors.append(f'{worker_id}_{i}_mismatch')
                    
                    time.sleep(0.001)  # Small delay to increase chance of race conditions
            except Exception as e:
                errors.append(f'{worker_id}_exception_{str(e)}')
        
        threads = []
        for i in range(5):
            thread = threading.Thread(target=worker, args=(i,))
            threads.append(thread)
            thread.start()
        
        for thread in threads:
            thread.join()
        
        assert len(errors) == 0, f"Errors occurred: {errors}"
        assert len(results) == 500  # 5 workers * 100 operations each
    
    def test_concurrent_scope_operations(self):
        """Test concurrent scope operations"""
        db = EphemeralDB()
        db.set('global_key', 'global_value')
        
        results = []
        errors = []
        
        def scope_worker(worker_id):
            try:
                with db.scope(f'worker_{worker_id}'):
                    db.set(f'local_key_{worker_id}', f'local_value_{worker_id}')
                    
                    # Verify we can see global data
                    global_val = db.get('global_key')
                    if global_val == 'global_value':
                        results.append(f'{worker_id}_global_success')
                    
                    # Verify we can see our local data
                    local_val = db.get(f'local_key_{worker_id}')
                    if local_val == f'local_value_{worker_id}':
                        results.append(f'{worker_id}_local_success')
                    
                    time.sleep(0.01)
            except Exception as e:
                errors.append(f'{worker_id}_exception_{str(e)}')
        
        threads = []
        for i in range(10):
            thread = threading.Thread(target=scope_worker, args=(i,))
            threads.append(thread)
            thread.start()
        
        for thread in threads:
            thread.join()
        
        assert len(errors) == 0, f"Errors occurred: {errors}"
        assert len(results) == 20  # 10 workers * 2 success checks each
        
        # Verify all scopes were cleaned up
        assert db.scope_count() == 1


class TestPerformance:
    """Basic performance tests"""
    
    def test_large_dataset_operations(self):
        """Test operations with large datasets"""
        db = EphemeralDB()
        
        # Set many keys
        start_time = time.time()
        for i in range(1000):
            db.set(f'key_{i}', f'value_{i}')
        set_time = time.time() - start_time
        
        # Get many keys
        start_time = time.time()
        for i in range(1000):
            value = db.get(f'key_{i}')
            assert value == f'value_{i}'
        get_time = time.time() - start_time
        
        # Basic performance assertions (should be fast)
        assert set_time < 1.0, f"Set operations took too long: {set_time}s"
        assert get_time < 1.0, f"Get operations took too long: {get_time}s"
    
    def test_deep_nesting_performance(self):
        """Test performance with deeply nested keys"""
        db = EphemeralDB()
        
        # Create deeply nested structure
        deep_key = '.'.join([f'level_{i}' for i in range(20)])
        
        start_time = time.time()
        db.set(deep_key, 'deep_value')
        set_time = time.time() - start_time
        
        start_time = time.time()
        value = db.get(deep_key)
        get_time = time.time() - start_time
        
        assert value == 'deep_value'
        assert set_time < 0.1, f"Deep set took too long: {set_time}s"
        assert get_time < 0.1, f"Deep get took too long: {get_time}s"


#!/usr/bin/env python3
"""
08. トランザクションシステム - ネストしたトランザクション処理
"""

from ephemeraldb import EphemeralDB
from typing import Dict, Any, List, Optional, Callable
from enum import Enum
import uuid
import time
from contextlib import contextmanager


class TransactionStatus(Enum):
    """トランザクション状態"""
    ACTIVE = "アクティブ"
    COMMITTED = "コミット済み"
    ABORTED = "中止"
    PREPARED = "準備完了"


class IsolationLevel(Enum):
    """分離レベル"""
    READ_UNCOMMITTED = "READ_UNCOMMITTED"
    READ_COMMITTED = "READ_COMMITTED"
    REPEATABLE_READ = "REPEATABLE_READ"
    SERIALIZABLE = "SERIALIZABLE"


class TransactionManager:
    """EphemeralDBを使用したトランザクション管理システム"""
    
    def __init__(self):
        self.db = EphemeralDB()
        self.active_transactions = {}
        self.transaction_stack = []
        
        # グローバルデータ領域の初期化
        self.db.set('グローバルデータ', {})
        self.db.set('ロック情報', {})
        self.db.set('トランザクション履歴', [])
    
    @contextmanager
    def transaction(self, name: Optional[str] = None, isolation_level: IsolationLevel = IsolationLevel.READ_COMMITTED):
        """トランザクションコンテキストマネージャー"""
        tx_id = str(uuid.uuid4())[:8]
        tx_name = name or f"TX_{tx_id}"
        
        # トランザクション開始
        self.begin_transaction(tx_id, tx_name, isolation_level)
        
        try:
            yield tx_id
            # 正常終了時はコミット
            self.commit_transaction(tx_id)
        except Exception as e:
            # 例外発生時はロールバック
            print(f"トランザクション例外: {e}")
            self.rollback_transaction(tx_id)
            raise
    
    def begin_transaction(self, tx_id: str, name: str, isolation_level: IsolationLevel):
        """トランザクション開始"""
        with self.db.scope(f'トランザクション_{tx_id}'):
            print(f"[{tx_id}] トランザクション開始: {name}")
            
            # トランザクション情報を設定
            self.db.set('トランザクションID', tx_id)
            self.db.set('トランザクション名', name)
            self.db.set('状態', TransactionStatus.ACTIVE)
            self.db.set('分離レベル', isolation_level)
            self.db.set('開始時刻', time.time())
            self.db.set('変更ログ', [])
            self.db.set('読み取りセット', set())
            self.db.set('書き込みセット', set())
            
            # アクティブトランザクションに登録
            self.active_transactions[tx_id] = {
                'name': name,
                'scope_count': self.db.scope_count(),
                'isolation_level': isolation_level
            }
            
            # トランザクションスタックにプッシュ
            self.transaction_stack.append(tx_id)
    
    def commit_transaction(self, tx_id: str):
        """トランザクションコミット"""
        if tx_id not in self.active_transactions:
            raise ValueError(f"アクティブでないトランザクション: {tx_id}")
        
        with self.db.scope(f'トランザクション_{tx_id}'):
            print(f"[{tx_id}] トランザクションコミット開始")
            
            # 変更ログを取得
            change_log = self.db.get('変更ログ', [])
            
            # 変更をグローバルデータに適用
            global_data = self.db.get('グローバルデータ')
            
            for change in change_log:
                if change['操作'] == 'SET':
                    self._set_global_data(global_data, change['キー'], change['新しい値'])
                elif change['操作'] == 'DELETE':
                    self._delete_global_data(global_data, change['キー'])
            
            self.db.set('グローバルデータ', global_data)
            
            # 状態を更新
            self.db.set('状態', TransactionStatus.COMMITTED)
            self.db.set('コミット時刻', time.time())
            
            # トランザクション履歴に記録
            history = self.db.get('トランザクション履歴')
            history.append({
                'トランザクションID': tx_id,
                '名前': self.db.get('トランザクション名'),
                '状態': TransactionStatus.COMMITTED.value,
                '開始時刻': self.db.get('開始時刻'),
                'コミット時刻': time.time(),
                '変更数': len(change_log)
            })
            self.db.set('トランザクション履歴', history)
            
            print(f"[{tx_id}] コミット完了 ({len(change_log)}件の変更)")
        
        # クリーンアップ
        self._cleanup_transaction(tx_id)
    
    def rollback_transaction(self, tx_id: str):
        """トランザクションロールバック"""
        if tx_id not in self.active_transactions:
            print(f"警告: アクティブでないトランザクション: {tx_id}")
            return
        
        with self.db.scope(f'トランザクション_{tx_id}'):
            print(f"[{tx_id}] トランザクションロールバック")
            
            # 状態を更新
            self.db.set('状態', TransactionStatus.ABORTED)
            self.db.set('ロールバック時刻', time.time())
            
            # 変更ログを取得（ログのみ、実際の変更は適用しない）
            change_log = self.db.get('変更ログ', [])
            
            # トランザクション履歴に記録
            history = self.db.get('トランザクション履歴')
            history.append({
                'トランザクションID': tx_id,
                '名前': self.db.get('トランザクション名'),
                '状態': TransactionStatus.ABORTED.value,
                '開始時刻': self.db.get('開始時刻'),
                'ロールバック時刻': time.time(),
                '変更数': len(change_log)
            })
            self.db.set('トランザクション履歴', history)
            
            print(f"[{tx_id}] ロールバック完了 ({len(change_log)}件の変更を破棄)")
        
        # クリーンアップ
        self._cleanup_transaction(tx_id)
    
    def set_data(self, tx_id: str, key: str, value: Any):
        """トランザクション内でデータ設定"""
        with self.db.scope(f'トランザクション_{tx_id}'):
            # 現在の値を取得（変更ログ用）
            current_value = self.get_data(tx_id, key)
            
            # 変更ログに記録
            change_log = self.db.get('変更ログ')
            change_log.append({
                '操作': 'SET',
                'キー': key,
                '古い値': current_value,
                '新しい値': value,
                '時刻': time.time()
            })
            self.db.set('変更ログ', change_log)
            
            # 書き込みセットに追加
            write_set = self.db.get('書き込みセット')
            write_set.add(key)
            self.db.set('書き込みセット', write_set)
            
            # トランザクション内の一時データとして保存
            self.db.set(f'データ.{key}', value)
            
            print(f"[{tx_id}] SET {key} = {value}")
    
    def get_data(self, tx_id: str, key: str, default: Any = None) -> Any:
        """トランザクション内でデータ取得"""
        with self.db.scope(f'トランザクション_{tx_id}'):
            # 読み取りセットに追加
            read_set = self.db.get('読み取りセット')
            read_set.add(key)
            self.db.set('読み取りセット', read_set)
            
            # まずトランザクション内の変更をチェック
            tx_value = self.db.get(f'データ.{key}')
            if tx_value is not None:
                return tx_value
            
            # グローバルデータから取得
            global_data = self.db.get('グローバルデータ')
            value = self._get_global_data(global_data, key, default)
            
            print(f"[{tx_id}] GET {key} => {value}")
            return value
    
    def delete_data(self, tx_id: str, key: str):
        """トランザクション内でデータ削除"""
        with self.db.scope(f'トランザクション_{tx_id}'):
            # 現在の値を取得（変更ログ用）
            current_value = self.get_data(tx_id, key)
            
            # 変更ログに記録
            change_log = self.db.get('変更ログ')
            change_log.append({
                '操作': 'DELETE',
                'キー': key,
                '古い値': current_value,
                '時刻': time.time()
            })
            self.db.set('変更ログ', change_log)
            
            # 書き込みセットに追加
            write_set = self.db.get('書き込みセット')
            write_set.add(key)
            self.db.set('書き込みセット', write_set)
            
            # トランザクション内で削除マーク
            self.db.set(f'データ.{key}', '__DELETED__')
            
            print(f"[{tx_id}] DELETE {key}")
    
    def _set_global_data(self, global_data: Dict, key: str, value: Any):
        """グローバルデータに値を設定（ドット記法対応）"""
        keys = key.split('.')
        current = global_data
        
        for k in keys[:-1]:
            if k not in current:
                current[k] = {}
            current = current[k]
        
        current[keys[-1]] = value
    
    def _get_global_data(self, global_data: Dict, key: str, default: Any = None) -> Any:
        """グローバルデータから値を取得（ドット記法対応）"""
        keys = key.split('.')
        current = global_data
        
        try:
            for k in keys:
                current = current[k]
            return current
        except (KeyError, TypeError):
            return default
    
    def _delete_global_data(self, global_data: Dict, key: str):
        """グローバルデータから値を削除（ドット記法対応）"""
        keys = key.split('.')
        current = global_data
        
        try:
            for k in keys[:-1]:
                current = current[k]
            if keys[-1] in current:
                del current[keys[-1]]
        except (KeyError, TypeError):
            pass
    
    def _cleanup_transaction(self, tx_id: str):
        """トランザクションクリーンアップ"""
        if tx_id in self.active_transactions:
            del self.active_transactions[tx_id]
        
        if tx_id in self.transaction_stack:
            self.transaction_stack.remove(tx_id)
    
    def get_transaction_status(self, tx_id: str) -> Optional[TransactionStatus]:
        """トランザクション状態取得"""
        if tx_id not in self.active_transactions:
            return None
        
        with self.db.scope(f'トランザクション_{tx_id}'):
            return self.db.get('状態')
    
    def get_active_transactions(self) -> List[str]:
        """アクティブなトランザクション一覧"""
        return list(self.active_transactions.keys())
    
    def get_transaction_history(self) -> List[Dict[str, Any]]:
        """トランザクション履歴取得"""
        return self.db.get('トランザクション履歴', [])


def basic_transaction_demo():
    """基本的なトランザクション機能のデモ"""
    print("=== 基本トランザクション機能 ===")
    
    tm = TransactionManager()
    
    # 初期データ設定
    with tm.transaction("初期データ設定") as tx_id:
        tm.set_data(tx_id, 'アカウント.Alice.残高', 1000)
        tm.set_data(tx_id, 'アカウント.Bob.残高', 500)
        tm.set_data(tx_id, 'アカウント.Charlie.残高', 750)
    
    print("初期データ設定完了")
    
    # 正常なトランザクション
    with tm.transaction("正常な送金") as tx_id:
        alice_balance = tm.get_data(tx_id, 'アカウント.Alice.残高')
        bob_balance = tm.get_data(tx_id, 'アカウント.Bob.残高')
        
        print(f"送金前 - Alice: {alice_balance}円, Bob: {bob_balance}円")
        
        # Alice から Bob に 200円送金
        transfer_amount = 200
        if alice_balance >= transfer_amount:
            tm.set_data(tx_id, 'アカウント.Alice.残高', alice_balance - transfer_amount)
            tm.set_data(tx_id, 'アカウント.Bob.残高', bob_balance + transfer_amount)
            
            new_alice = tm.get_data(tx_id, 'アカウント.Alice.残高')
            new_bob = tm.get_data(tx_id, 'アカウント.Bob.残高')
            print(f"送金後 - Alice: {new_alice}円, Bob: {new_bob}円")
    
    # 失敗するトランザクション（例外による自動ロールバック）
    try:
        with tm.transaction("失敗する送金") as tx_id:
            alice_balance = tm.get_data(tx_id, 'アカウント.Alice.残高')
            charlie_balance = tm.get_data(tx_id, 'アカウント.Charlie.残高')
            
            print(f"失敗送金前 - Alice: {alice_balance}円, Charlie: {charlie_balance}円")
            
            # Alice から Charlie に 2000円送金（残高不足）
            transfer_amount = 2000
            if alice_balance < transfer_amount:
                raise ValueError(f"残高不足: 必要{transfer_amount}円, 残高{alice_balance}円")
            
            tm.set_data(tx_id, 'アカウント.Alice.残高', alice_balance - transfer_amount)
            tm.set_data(tx_id, 'アカウント.Charlie.残高', charlie_balance + transfer_amount)
    
    except ValueError as e:
        print(f"送金失敗: {e}")
    
    # 最終残高確認（失敗したトランザクションは反映されない）
    with tm.transaction("残高確認") as tx_id:
        alice_final = tm.get_data(tx_id, 'アカウント.Alice.残高')
        bob_final = tm.get_data(tx_id, 'アカウント.Bob.残高')
        charlie_final = tm.get_data(tx_id, 'アカウント.Charlie.残高')
        
        print(f"最終残高 - Alice: {alice_final}円, Bob: {bob_final}円, Charlie: {charlie_final}円")
    
    print()


def nested_transaction_demo():
    """ネストしたトランザクションのデモ"""
    print("=== ネストしたトランザクション ===")
    
    tm = TransactionManager()
    
    # 複合的な業務処理のシミュレーション
    with tm.transaction("注文処理") as outer_tx:
        # 注文情報設定
        tm.set_data(outer_tx, '注文.ID', 'ORDER_001')
        tm.set_data(outer_tx, '注文.顧客', 'Alice')
        tm.set_data(outer_tx, '注文.商品', [{'商品ID': 'P001', '数量': 2, '単価': 500}])
        
        print("注文情報を設定しました")
        
        # 在庫チェックと更新（ネストしたトランザクション）
        try:
            with tm.transaction("在庫処理") as stock_tx:
                current_stock = tm.get_data(stock_tx, '在庫.P001', 10)
                required_qty = 2
                
                print(f"在庫チェック - 商品P001: 現在{current_stock}個, 必要{required_qty}個")
                
                if current_stock >= required_qty:
                    new_stock = current_stock - required_qty
                    tm.set_data(stock_tx, '在庫.P001', new_stock)
                    tm.set_data(stock_tx, '注文.在庫確保', True)
                    print(f"在庫を{required_qty}個確保。残り{new_stock}個")
                else:
                    raise ValueError("在庫不足")
        
        except ValueError as e:
            print(f"在庫処理失敗: {e}")
            # 外側のトランザクションも失敗させる
            raise
        
        # 決済処理（別のネストしたトランザクション）
        try:
            with tm.transaction("決済処理") as payment_tx:
                customer_balance = tm.get_data(payment_tx, 'アカウント.Alice.残高', 0)
                order_total = 1000  # 2 * 500円
                
                print(f"決済処理 - 顧客残高: {customer_balance}円, 注文金額: {order_total}円")
                
                if customer_balance >= order_total:
                    new_balance = customer_balance - order_total
                    tm.set_data(payment_tx, 'アカウント.Alice.残高', new_balance)
                    tm.set_data(payment_tx, '注文.決済完了', True)
                    tm.set_data(payment_tx, '注文.決済金額', order_total)
                    print(f"決済完了。残高: {new_balance}円")
                else:
                    raise ValueError("残高不足")
        
        except ValueError as e:
            print(f"決済処理失敗: {e}")
            raise
        
        # 注文完了
        tm.set_data(outer_tx, '注文.状態', '完了')
        tm.set_data(outer_tx, '注文.完了時刻', time.time())
        
        print("注文処理が正常に完了しました")
    
    print()


def concurrent_transaction_simulation():
    """同時実行トランザクションのシミュレーション"""
    print("=== 同時実行トランザクションシミュレーション ===")
    
    tm = TransactionManager()
    
    # 初期データ
    with tm.transaction("初期設定") as tx_id:
        tm.set_data(tx_id, '共有リソース.カウンター', 0)
        tm.set_data(tx_id, '共有リソース.合計', 0)
    
    # 複数のトランザクションを順次実行（実際の同時実行ではなく、概念的なデモ）
    transactions = []
    
    # トランザクション1: カウンターを増加
    print("トランザクション1開始（カウンター増加）")
    tm.begin_transaction("TX1", "カウンター増加", IsolationLevel.READ_COMMITTED)
    
    with tm.db.scope('トランザクション_TX1'):
        current_counter = tm.get_data("TX1", '共有リソース.カウンター')
        print(f"TX1: 現在のカウンター = {current_counter}")
        tm.set_data("TX1", '共有リソース.カウンター', current_counter + 1)
        
        # まだコミットしない（他のトランザクションの動作を確認するため）
        print("TX1: カウンターを+1（未コミット）")
    
    # トランザクション2: 合計を更新
    print("トランザクション2開始（合計更新）")
    tm.begin_transaction("TX2", "合計更新", IsolationLevel.READ_COMMITTED)
    
    with tm.db.scope('トランザクション_TX2'):
        # TX1の変更はまだコミットされていないので見えない
        current_counter = tm.get_data("TX2", '共有リソース.カウンター')
        current_total = tm.get_data("TX2", '共有リソース.合計')
        
        print(f"TX2: カウンター = {current_counter}, 合計 = {current_total}")
        new_total = current_total + current_counter * 10
        tm.set_data("TX2", '共有リソース.合計', new_total)
        print(f"TX2: 合計を{new_total}に更新")
    
    # TX2をコミット
    tm.commit_transaction("TX2")
    
    # TX1をコミット
    tm.commit_transaction("TX1")
    
    # 最終状態確認
    with tm.transaction("最終確認") as tx_id:
        final_counter = tm.get_data(tx_id, '共有リソース.カウンター')
        final_total = tm.get_data(tx_id, '共有リソース.合計')
        print(f"最終状態 - カウンター: {final_counter}, 合計: {final_total}")
    
    print()


def transaction_history_demo():
    """トランザクション履歴のデモ"""
    print("=== トランザクション履歴 ===")
    
    tm = TransactionManager()
    
    # 複数のトランザクションを実行
    with tm.transaction("ユーザー登録") as tx_id:
        tm.set_data(tx_id, 'ユーザー.user001.名前', '田中太郎')
        tm.set_data(tx_id, 'ユーザー.user001.メール', 'tanaka@example.com')
        tm.set_data(tx_id, 'ユーザー.user001.登録日', time.time())
    
    try:
        with tm.transaction("無効なデータ更新") as tx_id:
            tm.set_data(tx_id, 'ユーザー.user001.名前', '')  # 空の名前
            raise ValueError("名前が空です")
    except ValueError:
        pass  # ロールバックされる
    
    with tm.transaction("プロフィール更新") as tx_id:
        tm.set_data(tx_id, 'ユーザー.user001.プロフィール.年齢', 30)
        tm.set_data(tx_id, 'ユーザー.user001.プロフィール.職業', 'エンジニア')
    
    # 履歴表示
    history = tm.get_transaction_history()
    print("トランザクション履歴:")
    for i, tx_record in enumerate(history, 1):
        print(f"  {i}. {tx_record['名前']} - {tx_record['状態']} ({tx_record['変更数']}件の変更)")
    
    # 最終データ確認
    with tm.transaction("データ確認") as tx_id:
        user_name = tm.get_data(tx_id, 'ユーザー.user001.名前')
        user_email = tm.get_data(tx_id, 'ユーザー.user001.メール')
        user_age = tm.get_data(tx_id, 'ユーザー.user001.プロフィール.年齢')
        user_job = tm.get_data(tx_id, 'ユーザー.user001.プロフィール.職業')
        
        print(f"\n最終データ:")
        print(f"  名前: {user_name}")
        print(f"  メール: {user_email}")
        print(f"  年齢: {user_age}")
        print(f"  職業: {user_job}")
    
    print()


if __name__ == "__main__":
    basic_transaction_demo()
    nested_transaction_demo()
    concurrent_transaction_simulation()
    transaction_history_demo()
    
    print("=== まとめ ===")
    print("EphemeralDBを使用することで、本格的なトランザクション")
    print("システムを実装できます。ネストしたトランザクション、")
    print("自動ロールバック、分離レベルなどの高度な機能も")
    print("スコープ管理により自然に表現できます。")


#!/usr/bin/env python3
"""
09. デバッグコンテキスト - デバッグ情報を管理するシステム
"""

from ephemeraldb import EphemeralDB
import time
import traceback
import sys
from typing import Any, Dict, List, Optional, Callable
from contextlib import contextmanager
from functools import wraps


class DebugLevel:
    """デバッグレベル定数"""
    TRACE = 0
    DEBUG = 1
    INFO = 2
    WARNING = 3
    ERROR = 4
    CRITICAL = 5


class DebugContext:
    """EphemeralDBを使用したデバッグコンテキスト管理"""
    
    def __init__(self, name: str = "アプリケーション"):
        self.db = EphemeralDB()
        self.name = name
        
        # グローバル設定
        self.db.set('アプリ名', name)
        self.db.set('開始時刻', time.time())
        self.db.set('デバッグレベル', DebugLevel.INFO)
        self.db.set('ログ履歴', [])
        self.db.set('エラー履歴', [])
        self.db.set('パフォーマンス統計', {})
        self.db.set('呼び出しスタック', [])
    
    def set_debug_level(self, level: int):
        """デバッグレベルを設定"""
        self.db.set('デバッグレベル', level)
        level_names = ['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
        level_name = level_names[level] if 0 <= level < len(level_names) else 'UNKNOWN'
        print(f"デバッグレベルを{level_name}に設定")
    
    @contextmanager
    def function_scope(self, function_name: str, **kwargs):
        """関数スコープのコンテキストマネージャー"""
        with self.db.scope(f'関数_{function_name}'):
            start_time = time.time()
            
            # 関数情報を設定
            self.db.set('関数名', function_name)
            self.db.set('開始時刻', start_time)
            self.db.set('引数', kwargs)
            self.db.set('ローカル変数', {})
            self.db.set('実行ログ', [])
            
            # 呼び出しスタックに追加
            call_stack = self.db.get('呼び出しスタック')
            call_stack.append({
                '関数名': function_name,
                '開始時刻': start_time,
                '引数': kwargs
            })
            self.db.set('呼び出しスタック', call_stack)
            
            self.log(DebugLevel.DEBUG, f"関数 {function_name} 開始", kwargs)
            
            try:
                yield self
                
                # 正常終了
                end_time = time.time()
                execution_time = end_time - start_time
                
                self.db.set('終了時刻', end_time)
                self.db.set('実行時間', execution_time)
                self.db.set('正常終了', True)
                
                self.log(DebugLevel.DEBUG, f"関数 {function_name} 正常終了", {
                    '実行時間': f'{execution_time:.4f}秒'
                })
                
                # パフォーマンス統計更新
                self._update_performance_stats(function_name, execution_time)
                
            except Exception as e:
                # 例外発生
                end_time = time.time()
                execution_time = end_time - start_time
                
                self.db.set('終了時刻', end_time)
                self.db.set('実行時間', execution_time)
                self.db.set('正常終了', False)
                self.db.set('例外', {
                    'タイプ': type(e).__name__,
                    'メッセージ': str(e),
                    'トレースバック': traceback.format_exc()
                })
                
                self.log(DebugLevel.ERROR, f"関数 {function_name} で例外発生", {
                    '例外タイプ': type(e).__name__,
                    'メッセージ': str(e),
                    '実行時間': f'{execution_time:.4f}秒'
                })
                
                # エラー履歴に記録
                self._record_error(function_name, e, execution_time)
                
                raise
            
            finally:
                # 呼び出しスタックから削除
                call_stack = self.db.get('呼び出しスタック')
                if call_stack:
                    call_stack.pop()
                    self.db.set('呼び出しスタック', call_stack)
    
    def log(self, level: int, message: str, data: Optional[Dict[str, Any]] = None):
        """ログメッセージを記録"""
        current_level = self.db.get('デバッグレベル')
        if level < current_level:
            return
        
        # ログエントリ作成
        log_entry = {
            '時刻': time.time(),
            'レベル': level,
            'メッセージ': message,
            'データ': data or {},
            'スコープ': self.db.current_scope(),
            'スタック深度': len(self.db.get('呼び出しスタック', []))
        }
        
        # 現在のスコープのログに追加
        execution_log = self.db.get('実行ログ', [])
        execution_log.append(log_entry)
        self.db.set('実行ログ', execution_log)
        
        # グローバルログ履歴に追加
        log_history = self.db.get('ログ履歴')
        log_history.append(log_entry)
        self.db.set('ログ履歴', log_history)
        
        # コンソール出力
        level_names = ['TRACE', 'DEBUG', 'INFO', 'WARN', 'ERROR', 'CRIT']
        level_name = level_names[level] if 0 <= level < len(level_names) else 'UNK'
        indent = '  ' * log_entry['スタック深度']
        
        print(f"{indent}[{level_name}] {message}")
        if data:
            for key, value in data.items():
                print(f"{indent}    {key}: {value}")
    
    def set_variable(self, name: str, value: Any):
        """ローカル変数を設定"""
        local_vars = self.db.get('ローカル変数', {})
        local_vars[name] = value
        self.db.set('ローカル変数', local_vars)
        
        self.log(DebugLevel.TRACE, f"変数設定: {name}", {'値': value})
    
    def get_variable(self, name: str, default: Any = None) -> Any:
        """ローカル変数を取得"""
        local_vars = self.db.get('ローカル変数', {})
        value = local_vars.get(name, default)
        
        self.log(DebugLevel.TRACE, f"変数取得: {name}", {'値': value})
        return value
    
    def checkpoint(self, name: str, data: Optional[Dict[str, Any]] = None):
        """チェックポイントを設定"""
        checkpoint_data = {
            '名前': name,
            '時刻': time.time(),
            'データ': data or {},
            'ローカル変数': self.db.get('ローカル変数', {}),
            'スコープ': self.db.current_scope()
        }
        
        checkpoints = self.db.get('チェックポイント', [])
        checkpoints.append(checkpoint_data)
        self.db.set('チェックポイント', checkpoints)
        
        self.log(DebugLevel.DEBUG, f"チェックポイント: {name}", data)
    
    def assert_condition(self, condition: bool, message: str, data: Optional[Dict[str, Any]] = None):
        """アサーション"""
        if not condition:
            self.log(DebugLevel.ERROR, f"アサーション失敗: {message}", data)
            raise AssertionError(message)
        else:
            self.log(DebugLevel.TRACE, f"アサーション成功: {message}", data)
    
    def _update_performance_stats(self, function_name: str, execution_time: float):
        """パフォーマンス統計更新"""
        stats = self.db.get('パフォーマンス統計')
        
        if function_name not in stats:
            stats[function_name] = {
                '実行回数': 0,
                '総実行時間': 0.0,
                '最小実行時間': float('inf'),
                '最大実行時間': 0.0
            }
        
        func_stats = stats[function_name]
        func_stats['実行回数'] += 1
        func_stats['総実行時間'] += execution_time
        func_stats['最小実行時間'] = min(func_stats['最小実行時間'], execution_time)
        func_stats['最大実行時間'] = max(func_stats['最大実行時間'], execution_time)
        func_stats['平均実行時間'] = func_stats['総実行時間'] / func_stats['実行回数']
        
        self.db.set('パフォーマンス統計', stats)
    
    def _record_error(self, function_name: str, exception: Exception, execution_time: float):
        """エラー記録"""
        error_record = {
            '時刻': time.time(),
            '関数名': function_name,
            '例外タイプ': type(exception).__name__,
            'メッセージ': str(exception),
            '実行時間': execution_time,
            'トレースバック': traceback.format_exc(),
            'スコープ': self.db.current_scope()
        }
        
        error_history = self.db.get('エラー履歴')
        error_history.append(error_record)
        self.db.set('エラー履歴', error_history)
    
    def get_performance_stats(self) -> Dict[str, Any]:
        """パフォーマンス統計取得"""
        return self.db.get('パフォーマンス統計', {})
    
    def get_error_history(self) -> List[Dict[str, Any]]:
        """エラー履歴取得"""
        return self.db.get('エラー履歴', [])
    
    def get_log_history(self) -> List[Dict[str, Any]]:
        """ログ履歴取得"""
        return self.db.get('ログ履歴', [])
    
    def print_summary(self):
        """デバッグサマリーを出力"""
        print("\n=== デバッグサマリー ===")
        
        # パフォーマンス統計
        stats = self.get_performance_stats()
        if stats:
            print("\nパフォーマンス統計:")
            for func_name, func_stats in stats.items():
                print(f"  {func_name}:")
                print(f"    実行回数: {func_stats['実行回数']}回")
                print(f"    平均実行時間: {func_stats['平均実行時間']:.4f}秒")
                print(f"    最小/最大: {func_stats['最小実行時間']:.4f}秒 / {func_stats['最大実行時間']:.4f}秒")
        
        # エラー統計
        errors = self.get_error_history()
        if errors:
            print(f"\nエラー統計: {len(errors)}件のエラー")
            error_types = {}
            for error in errors:
                error_type = error['例外タイプ']
                error_types[error_type] = error_types.get(error_type, 0) + 1
            
            for error_type, count in error_types.items():
                print(f"  {error_type}: {count}件")
        
        # ログ統計
        logs = self.get_log_history()
        if logs:
            print(f"\nログ統計: {len(logs)}件のログ")
            level_counts = {}
            level_names = ['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
            
            for log in logs:
                level = log['レベル']
                level_name = level_names[level] if 0 <= level < len(level_names) else 'UNKNOWN'
                level_counts[level_name] = level_counts.get(level_name, 0) + 1
            
            for level_name, count in level_counts.items():
                print(f"  {level_name}: {count}件")


def debug_decorator(debug_context: DebugContext):
    """デバッグ用デコレータ"""
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            with debug_context.function_scope(func.__name__, 引数=args, キーワード引数=kwargs):
                return func(*args, **kwargs)
        return wrapper
    return decorator


def basic_debugging_demo():
    """基本的なデバッグ機能のデモ"""
    print("=== 基本デバッグ機能 ===")
    
    debug = DebugContext("計算アプリ")
    debug.set_debug_level(DebugLevel.DEBUG)
    
    @debug_decorator(debug)
    def add_numbers(a: int, b: int) -> int:
        debug.set_variable('a', a)
        debug.set_variable('b', b)
        
        debug.checkpoint("引数チェック", {'a': a, 'b': b})
        debug.assert_condition(isinstance(a, (int, float)), "aは数値である必要があります")
        debug.assert_condition(isinstance(b, (int, float)), "bは数値である必要があります")
        
        result = a + b
        debug.set_variable('result', result)
        
        debug.checkpoint("計算完了", {'result': result})
        debug.log(DebugLevel.INFO, "加算処理完了", {'結果': result})
        
        return result
    
    @debug_decorator(debug)
    def multiply_numbers(a: int, b: int) -> int:
        debug.set_variable('a', a)
        debug.set_variable('b', b)
        
        # add_numbers を呼び出し（ネストしたデバッグコンテキスト）
        temp_result = add_numbers(a, 0)  # a + 0 = a
        
        result = a * b
        debug.set_variable('result', result)
        
        debug.log(DebugLevel.INFO, "乗算処理完了", {'結果': result})
        return result
    
    # 正常なケース
    result1 = add_numbers(10, 20)
    print(f"加算結果: {result1}")
    
    result2 = multiply_numbers(5, 4)
    print(f"乗算結果: {result2}")
    
    # エラーケース
    try:
        add_numbers("文字列", 20)
    except AssertionError as e:
        print(f"エラーをキャッチ: {e}")
    
    debug.print_summary()
    print()


def complex_algorithm_debugging():
    """複雑なアルゴリズムのデバッグ"""
    print("=== 複雑なアルゴリズムのデバッグ ===")
    
    debug = DebugContext("ソートアルゴリズム")
    debug.set_debug_level(DebugLevel.INFO)
    
    @debug_decorator(debug)
    def bubble_sort(arr: List[int]) -> List[int]:
        debug.set_variable('original_array', arr.copy())
        debug.set_variable('array_length', len(arr))
        
        arr = arr.copy()  # 元の配列を変更しない
        n = len(arr)
        
        debug.checkpoint("ソート開始", {'配列長': n, '元の配列': arr})
        
        for i in range(n):
            debug.log(DebugLevel.DEBUG, f"外側ループ {i+1}/{n}")
            
            with debug.db.scope(f'外側ループ_{i}'):
                debug.set_variable('outer_index', i)
                swapped = False
                
                for j in range(0, n - i - 1):
                    with debug.db.scope(f'内側ループ_{j}'):
                        debug.set_variable('inner_index', j)
                        debug.set_variable('comparing', [arr[j], arr[j + 1]])
                        
                        if arr[j] > arr[j + 1]:
                            # スワップ
                            debug.log(DebugLevel.TRACE, f"スワップ: {arr[j]} <-> {arr[j + 1]}")
                            arr[j], arr[j + 1] = arr[j + 1], arr[j]
                            swapped = True
                            
                            debug.checkpoint("スワップ実行", {
                                'インデックス': [j, j + 1],
                                '現在の配列': arr.copy()
                            })
                
                debug.set_variable('swapped_in_pass', swapped)
                
                if not swapped:
                    debug.log(DebugLevel.INFO, f"第{i+1}パスでスワップなし - ソート完了")
                    break
                
                debug.checkpoint(f"第{i+1}パス完了", {'配列状態': arr.copy()})
        
        debug.checkpoint("ソート完了", {'最終配列': arr})
        debug.log(DebugLevel.INFO, "バブルソート完了", {'結果': arr})
        
        return arr
    
    @debug_decorator(debug)
    def find_max(arr: List[int]) -> Dict[str, Any]:
        debug.set_variable('array', arr)
        debug.assert_condition(len(arr) > 0, "配列が空です")
        
        max_value = arr[0]
        max_index = 0
        
        debug.checkpoint("最大値検索開始", {'初期値': max_value})
        
        for i, value in enumerate(arr[1:], 1):
            with debug.db.scope(f'要素_{i}'):
                debug.set_variable('current_value', value)
                debug.set_variable('current_max', max_value)
                
                if value > max_value:
                    debug.log(DebugLevel.DEBUG, f"新しい最大値発見: {value} > {max_value}")
                    max_value = value
                    max_index = i
                    
                    debug.checkpoint("最大値更新", {
                        '新しい最大値': max_value,
                        'インデックス': max_index
                    })
        
        result = {'値': max_value, 'インデックス': max_index}
        debug.log(DebugLevel.INFO, "最大値検索完了", result)
        
        return result
    
    # テストデータ
    test_array = [64, 34, 25, 12, 22, 11, 90]
    
    print(f"元の配列: {test_array}")
    
    # 最大値検索
    max_info = find_max(test_array)
    print(f"最大値: {max_info['値']} (インデックス: {max_info['インデックス']})")
    
    # ソート実行
    sorted_array = bubble_sort(test_array)
    print(f"ソート後: {sorted_array}")
    
    debug.print_summary()
    print()


def error_handling_debugging():
    """エラーハンドリングのデバッグ"""
    print("=== エラーハンドリングのデバッグ ===")
    
    debug = DebugContext("エラーハンドリング")
    debug.set_debug_level(DebugLevel.DEBUG)
    
    @debug_decorator(debug)
    def risky_operation(value: Any) -> str:
        debug.set_variable('input_value', value)
        
        debug.checkpoint("入力検証開始")
        
        if value is None:
            debug.log(DebugLevel.WARNING, "None値が渡されました")
            raise ValueError("None値は処理できません")
        
        if isinstance(value, str) and len(value) == 0:
            debug.log(DebugLevel.WARNING, "空文字列が渡されました")
            raise ValueError("空文字列は処理できません")
        
        debug.checkpoint("入力検証完了")
        
        # 型に応じた処理
        if isinstance(value, (int, float)):
            with debug.db.scope('数値処理'):
                debug.log(DebugLevel.DEBUG, "数値として処理")
                
                if value < 0:
                    debug.log(DebugLevel.WARNING, "負の数値です")
                    raise ValueError("負の数値は処理できません")
                
                result = f"数値: {value * 2}"
                debug.checkpoint("数値処理完了", {'結果': result})
                
        elif isinstance(value, str):
            with debug.db.scope('文字列処理'):
                debug.log(DebugLevel.DEBUG, "文字列として処理")
                
                if len(value) > 10:
                    debug.log(DebugLevel.WARNING, "文字列が長すぎます")
                    raise ValueError("文字列が長すぎます（10文字以下にしてください）")
                
                result = f"文字列: {value.upper()}"
                debug.checkpoint("文字列処理完了", {'結果': result})
                
        else:
            debug.log(DebugLevel.ERROR, f"サポートされていない型: {type(value)}")
            raise TypeError(f"型 {type(value)} はサポートされていません")
        
        debug.log(DebugLevel.INFO, "処理完了", {'結果': result})
        return result
    
    # 様々なテストケース
    test_cases = [
        (5, "正常な数値"),
        ("hello", "正常な文字列"),
        (-3, "負の数値（エラー）"),
        ("", "空文字列（エラー）"),
        ("very long string that exceeds limit", "長すぎる文字列（エラー）"),
        (None, "None値（エラー）"),
        ([1, 2, 3], "リスト（エラー）")
    ]
    
    for value, description in test_cases:
        print(f"\nテスト: {description}")
        try:
            result = risky_operation(value)
            print(f"結果: {result}")
        except Exception as e:
            print(f"例外: {type(e).__name__}: {e}")
    
    debug.print_summary()
    print()


def performance_debugging():
    """パフォーマンスのデバッグ"""
    print("=== パフォーマンスのデバッグ ===")
    
    debug = DebugContext("パフォーマンステスト")
    debug.set_debug_level(DebugLevel.INFO)
    
    @debug_decorator(debug)
    def fibonacci_recursive(n: int) -> int:
        debug.set_variable('n', n)
        
        if n <= 1:
            debug.log(DebugLevel.TRACE, f"ベースケース: fibonacci({n}) = {n}")
            return n
        
        debug.checkpoint(f"再帰計算開始", {'n': n})
        
        # 再帰呼び出し
        fib_n1 = fibonacci_recursive(n - 1)
        fib_n2 = fibonacci_recursive(n - 2)
        
        result = fib_n1 + fib_n2
        debug.log(DebugLevel.TRACE, f"fibonacci({n}) = {fib_n1} + {fib_n2} = {result}")
        
        return result
    
    @debug_decorator(debug)
    def fibonacci_iterative(n: int) -> int:
        debug.set_variable('n', n)
        
        if n <= 1:
            return n
        
        a, b = 0, 1
        debug.set_variable('a', a)
        debug.set_variable('b', b)
        
        for i in range(2, n + 1):
            with debug.db.scope(f'反復_{i}'):
                debug.set_variable('iteration', i)
                a, b = b, a + b
                debug.log(DebugLevel.TRACE, f"fibonacci({i}) = {b}")
        
        debug.log(DebugLevel.DEBUG, f"反復計算完了: fibonacci({n}) = {b}")
        return b
    
    # パフォーマンス比較
    test_values = [5, 10, 15]
    
    for n in test_values:
        print(f"\nフィボナッチ数列の第{n}項を計算:")
        
        # 再帰版
        result_recursive = fibonacci_recursive(n)
        print(f"再帰版結果: {result_recursive}")
        
        # 反復版
        result_iterative = fibonacci_iterative(n)
        print(f"反復版結果: {result_iterative}")
    
    debug.print_summary()
    print()


if __name__ == "__main__":
    basic_debugging_demo()
    complex_algorithm_debugging()
    error_handling_debugging()
    performance_debugging()
    
    print("=== まとめ ===")
    print("EphemeralDBを使用することで、デバッグ情報を")
    print("階層的に管理し、詳細な実行トレースを取得できます。")
    print("関数スコープ、変数追跡、パフォーマンス測定など")
    print("開発に必要なデバッグ機能を効率的に実装できます。")


#!/usr/bin/env python3
"""
04. DSLインタープリター - EphemeralDBを使用したDSL実装
"""

from ephemeraldb import EphemeralDB
from typing import List, Dict, Any, Union


class SimpleDSLInterpreter:
    """EphemeralDBを使用したシンプルなDSLインタープリター"""
    
    def __init__(self):
        self.db = EphemeralDB()
        self.functions = {
            'SET': self._set_variable,
            'GET': self._get_variable,
            'PRINT': self._print_value,
            'BLOCK': self._execute_block,
            'IF': self._execute_if,
            'WHILE': self._execute_while,
            'FUNCTION': self._define_function,
            'CALL': self._call_function,
            'ADD': self._add_numbers,
            'MULTIPLY': self._multiply_numbers,
        }
        self.user_functions = {}
    
    def execute(self, commands: List[Dict[str, Any]]) -> List[Any]:
        """DSLコマンドのリストを実行"""
        results = []
        
        for command in commands:
            cmd_type = command.get('type')
            if cmd_type in self.functions:
                result = self.functions[cmd_type](command)
                if result is not None:
                    results.append(result)
            else:
                raise ValueError(f"不明なコマンド型: {cmd_type}")
        
        return results
    
    def _set_variable(self, command: Dict[str, Any]) -> None:
        """SET 変数 値"""
        var_name = command['variable']
        value = self._evaluate_expression(command['value'])
        self.db.set(var_name, value)
        print(f"[SET] {var_name} = {value}")
    
    def _get_variable(self, command: Dict[str, Any]) -> Any:
        """GET 変数"""
        var_name = command['variable']
        value = self.db.get(var_name)
        print(f"[GET] {var_name} => {value}")
        return value
    
    def _print_value(self, command: Dict[str, Any]) -> str:
        """PRINT 式"""
        value = self._evaluate_expression(command['expression'])
        print(f"[出力] {value}")
        return str(value)
    
    def _execute_block(self, command: Dict[str, Any]) -> List[Any]:
        """BLOCK - 新しいスコープでコマンドを実行"""
        block_name = command.get('name', '無名ブロック')
        
        with self.db.scope(block_name):
            print(f"[ブロック開始] {block_name}")
            
            # ローカル変数があればインポート
            local_vars = command.get('locals', {})
            for var, value in local_vars.items():
                self.db.set(var, self._evaluate_expression(value))
                print(f"[ローカル変数] {var} = {value}")
            
            results = self.execute(command['commands'])
            print(f"[ブロック終了] {block_name}")
            return results
    
    def _execute_if(self, command: Dict[str, Any]) -> Any:
        """IF 条件 then_commands [else_commands]"""
        condition = self._evaluate_expression(command['condition'])
        print(f"[IF] 条件: {condition}")
        
        if condition:
            print("[IF] then分岐を実行")
            return self.execute(command['then_commands'])
        elif 'else_commands' in command:
            print("[IF] else分岐を実行")
            return self.execute(command['else_commands'])
        else:
            print("[IF] 実行する分岐なし")
    
    def _execute_while(self, command: Dict[str, Any]) -> List[Any]:
        """WHILE 条件 commands"""
        results = []
        max_iterations = command.get('max_iterations', 10)  # 安全のための上限
        
        iteration = 0
        print(f"[WHILE] ループ開始（最大{max_iterations}回）")
        
        while iteration < max_iterations:
            condition = self._evaluate_expression(command['condition'])
            print(f"[WHILE] 反復{iteration}: 条件 = {condition}")
            
            if not condition:
                break
            
            # 各反復を新しいスコープで実行
            with self.db.scope(f'while_反復_{iteration}'):
                self.db.set('_反復回数', iteration)
                results.extend(self.execute(command['commands']))
            
            iteration += 1
        
        print(f"[WHILE] ループ終了（{iteration}回実行）")
        return results
    
    def _define_function(self, command: Dict[str, Any]) -> None:
        """FUNCTION 名前 パラメータ commands"""
        func_name = command['name']
        self.user_functions[func_name] = {
            'parameters': command.get('parameters', []),
            'commands': command['commands']
        }
        print(f"[関数定義] {func_name} ({', '.join(command.get('parameters', []))})")
    
    def _call_function(self, command: Dict[str, Any]) -> Any:
        """CALL 関数名 引数"""
        func_name = command['function']
        
        if func_name not in self.user_functions:
            raise ValueError(f"未定義の関数: {func_name}")
        
        func_def = self.user_functions[func_name]
        arguments = [self._evaluate_expression(arg) for arg in command.get('arguments', [])]
        
        print(f"[関数呼び出し] {func_name}({', '.join(map(str, arguments))})")
        
        # 関数を新しいスコープで実行
        with self.db.scope(f'関数_{func_name}'):
            # パラメータをバインド
            for param, arg in zip(func_def['parameters'], arguments):
                self.db.set(param, arg)
                print(f"[パラメータ] {param} = {arg}")
            
            results = self.execute(func_def['commands'])
            result = results[-1] if results else None
            print(f"[関数結果] {func_name} => {result}")
            return result
    
    def _add_numbers(self, command: Dict[str, Any]) -> float:
        """ADD 数値1 数値2"""
        a = self._evaluate_expression(command['a'])
        b = self._evaluate_expression(command['b'])
        result = a + b
        print(f"[ADD] {a} + {b} = {result}")
        return result
    
    def _multiply_numbers(self, command: Dict[str, Any]) -> float:
        """MULTIPLY 数値1 数値2"""
        a = self._evaluate_expression(command['a'])
        b = self._evaluate_expression(command['b'])
        result = a * b
        print(f"[MULTIPLY] {a} * {b} = {result}")
        return result
    
    def _evaluate_expression(self, expression: Union[str, int, float, Dict]) -> Any:
        """式を評価（簡略化版）"""
        if isinstance(expression, (int, float, str, bool)):
            return expression
        elif isinstance(expression, dict):
            if expression.get('type') == 'variable':
                return self.db.get(expression['name'])
            elif expression.get('type') == 'operation':
                return self._evaluate_operation(expression)
        
        return expression
    
    def _evaluate_operation(self, operation: Dict[str, Any]) -> Any:
        """数学的/論理的操作を評価"""
        op = operation['operator']
        left = self._evaluate_expression(operation['left'])
        right = self._evaluate_expression(operation['right'])
        
        operations = {
            '+': lambda a, b: a + b,
            '-': lambda a, b: a - b,
            '*': lambda a, b: a * b,
            '/': lambda a, b: a / b if b != 0 else 0,
            '==': lambda a, b: a == b,
            '!=': lambda a, b: a != b,
            '<': lambda a, b: a < b,
            '>': lambda a, b: a > b,
            '<=': lambda a, b: a <= b,
            '>=': lambda a, b: a >= b,
            'and': lambda a, b: a and b,
            'or': lambda a, b: a or b,
        }
        
        if op in operations:
            result = operations[op](left, right)
            print(f"[演算] {left} {op} {right} = {result}")
            return result
        else:
            raise ValueError(f"未知の演算子: {op}")


def basic_variables_demo():
    """基本的な変数操作のデモ"""
    print("=== 基本変数操作 ===")
    
    interpreter = SimpleDSLInterpreter()
    
    commands = [
        {'type': 'SET', 'variable': '名前', 'value': 'アリス'},
        {'type': 'SET', 'variable': '年齢', 'value': 30},
        {'type': 'PRINT', 'expression': {'type': 'variable', 'name': '名前'}},
        {'type': 'PRINT', 'expression': {'type': 'variable', 'name': '年齢'}},
        {'type': 'GET', 'variable': '名前'},
    ]
    
    interpreter.execute(commands)
    print()


def scope_demo():
    """スコープ機能のデモ"""
    print("=== スコープ機能 ===")
    
    interpreter = SimpleDSLInterpreter()
    
    commands = [
        {'type': 'SET', 'variable': 'グローバル変数', 'value': 'グローバルです'},
        {'type': 'BLOCK', 'name': '内部スコープ', 
         'locals': {'ローカル変数': 'ローカルです'}, 
         'commands': [
             {'type': 'PRINT', 'expression': {'type': 'variable', 'name': 'グローバル変数'}},
             {'type': 'PRINT', 'expression': {'type': 'variable', 'name': 'ローカル変数'}},
             {'type': 'SET', 'variable': '一時変数', 'value': '一時的'},
             {'type': 'PRINT', 'expression': {'type': 'variable', 'name': '一時変数'}},
         ]},
        {'type': 'PRINT', 'expression': {'type': 'variable', 'name': 'グローバル変数'}},
        {'type': 'PRINT', 'expression': {'type': 'variable', 'name': '一時変数'}},  # None
    ]
    
    interpreter.execute(commands)
    print()


def conditional_demo():
    """条件分岐のデモ"""
    print("=== 条件分岐 ===")
    
    interpreter = SimpleDSLInterpreter()
    
    commands = [
        {'type': 'SET', 'variable': 'x', 'value': 10},
        {'type': 'SET', 'variable': 'y', 'value': 5},
        {'type': 'IF', 
         'condition': {'type': 'operation', 'operator': '>', 
                      'left': {'type': 'variable', 'name': 'x'}, 
                      'right': {'type': 'variable', 'name': 'y'}},
         'then_commands': [
             {'type': 'PRINT', 'expression': 'xはyより大きいです'}
         ],
         'else_commands': [
             {'type': 'PRINT', 'expression': 'xはy以下です'}
         ]
        }
    ]
    
    interpreter.execute(commands)
    print()


def function_demo():
    """関数定義と呼び出しのデモ"""
    print("=== 関数定義と呼び出し ===")
    
    interpreter = SimpleDSLInterpreter()
    
    commands = [
        # 関数を定義
        {'type': 'FUNCTION', 'name': '数値を2倍', 'parameters': ['数値'], 'commands': [
            {'type': 'MULTIPLY', 'a': {'type': 'variable', 'name': '数値'}, 'b': 2},
        ]},
        
        # 関数を呼び出し
        {'type': 'CALL', 'function': '数値を2倍', 'arguments': [21]},
        
        # より複雑な関数
        {'type': 'FUNCTION', 'name': '計算と表示', 'parameters': ['a', 'b'], 'commands': [
            {'type': 'ADD', 'a': {'type': 'variable', 'name': 'a'}, 'b': {'type': 'variable', 'name': 'b'}},
            {'type': 'SET', 'variable': '結果', 'value': {'type': 'operation', 'operator': '+', 
                     'left': {'type': 'variable', 'name': 'a'}, 
                     'right': {'type': 'variable', 'name': 'b'}}},
            {'type': 'PRINT', 'expression': {'type': 'variable', 'name': '結果'}},
            {'type': 'GET', 'variable': '結果'},
        ]},
        
        {'type': 'CALL', 'function': '計算と表示', 'arguments': [15, 25]},
    ]
    
    results = interpreter.execute(commands)
    print(f"関数の結果: {results}")
    print()


def loop_demo():
    """ループのデモ"""
    print("=== ループ処理 ===")
    
    interpreter = SimpleDSLInterpreter()
    
    commands = [
        {'type': 'SET', 'variable': 'カウンター', 'value': 0},
        {'type': 'WHILE', 
         'condition': {'type': 'operation', 'operator': '<', 
                      'left': {'type': 'variable', 'name': 'カウンター'}, 
                      'right': 3},
         'max_iterations': 5,
         'commands': [
             {'type': 'PRINT', 'expression': {'type': 'variable', 'name': 'カウンター'}},
             {'type': 'SET', 'variable': 'ループローカル', 
              'value': {'type': 'operation', 'operator': '*', 
                       'left': {'type': 'variable', 'name': 'カウンター'}, 
                       'right': 10}},
             {'type': 'PRINT', 'expression': {'type': 'variable', 'name': 'ループローカル'}},
             {'type': 'SET', 'variable': 'カウンター', 
              'value': {'type': 'operation', 'operator': '+', 
                       'left': {'type': 'variable', 'name': 'カウンター'}, 
                       'right': 1}},
         ]
        },
        {'type': 'PRINT', 'expression': 'ループ終了'},
        {'type': 'GET', 'variable': 'カウンター'},
        {'type': 'GET', 'variable': 'ループローカル'},  # None（スコープ外）
    ]
    
    interpreter.execute(commands)
    print()


def complex_program_demo():
    """複雑なプログラムのデモ"""
    print("=== 複雑なプログラム ===")
    
    interpreter = SimpleDSLInterpreter()
    
    # 階乗を計算するDSLプログラム
    commands = [
        # 階乗関数を定義（簡略版）
        {'type': 'FUNCTION', 'name': '階乗計算', 'parameters': ['n'], 'commands': [
            {'type': 'SET', 'variable': '結果', 'value': 1},
            {'type': 'SET', 'variable': 'i', 'value': 1},
            {'type': 'WHILE', 
             'condition': {'type': 'operation', 'operator': '<=', 
                          'left': {'type': 'variable', 'name': 'i'}, 
                          'right': {'type': 'variable', 'name': 'n'}},
             'max_iterations': 10,
             'commands': [
                 {'type': 'SET', 'variable': '結果', 
                  'value': {'type': 'operation', 'operator': '*', 
                           'left': {'type': 'variable', 'name': '結果'}, 
                           'right': {'type': 'variable', 'name': 'i'}}},
                 {'type': 'SET', 'variable': 'i', 
                  'value': {'type': 'operation', 'operator': '+', 
                           'left': {'type': 'variable', 'name': 'i'}, 
                           'right': 1}},
             ]
            },
            {'type': 'GET', 'variable': '結果'},
        ]},
        
        # メインプログラム
        {'type': 'BLOCK', 'name': 'メインプログラム', 'commands': [
            {'type': 'PRINT', 'expression': '階乗計算プログラム開始'},
            {'type': 'SET', 'variable': '入力', 'value': 5},
            {'type': 'PRINT', 'expression': {'type': 'variable', 'name': '入力'}},
            {'type': 'CALL', 'function': '階乗計算', 'arguments': [{'type': 'variable', 'name': '入力'}]},
            {'type': 'PRINT', 'expression': '計算完了'},
        ]},
    ]
    
    results = interpreter.execute(commands)
    print(f"最終結果: {results}")
    print()


if __name__ == "__main__":
    basic_variables_demo()
    scope_demo()
    conditional_demo()
    function_demo()
    loop_demo()
    complex_program_demo()
    
    print("=== まとめ ===")
    print("EphemeralDBにより、DSLインタープリターで")
    print("変数スコープを簡単に管理できます。")
    print("関数呼び出し、ループ、条件分岐すべてで")
    print("適切な変数の分離が実現されます。")